import warnings
from dataclasses import dataclass
from functools import wraps
from string import punctuation
from typing import Callable, Iterable, Literal, Sequence, Tuple, overload

import torch
from transformers import BatchEncoding
from transformers.activations import ACT2FN

from ..base import LightningIRModel, LightningIROutput
from ..base.model import _batch_encoding
from . import BiEncoderConfig


class MLMHead(torch.nn.Module):
    def __init__(self, config: BiEncoderConfig) -> None:
        super().__init__()
        self.config = config
        self.dense = torch.nn.Linear(config.hidden_size, config.hidden_size)
        if isinstance(config.hidden_act, str):
            self.transform_act_fn = ACT2FN[config.hidden_act]
        else:
            self.transform_act_fn = config.hidden_act
        self.LayerNorm = torch.nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.decoder = torch.nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        self.bias = torch.nn.Parameter(torch.zeros(config.vocab_size))

        self.decoder.bias = self.bias

    def _tie_weights(self):
        self.decoder.bias = self.bias

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.transform_act_fn(hidden_states)
        hidden_states = self.LayerNorm(hidden_states)
        hidden_states = self.decoder(hidden_states)
        return hidden_states


@dataclass
class BiEncoderEmbedding:
    embeddings: torch.Tensor
    scoring_mask: torch.Tensor

    @overload
    def to(self, device: torch.device, /) -> "BiEncoderEmbedding": ...

    @overload
    def to(self, other: "BiEncoderEmbedding", /) -> "BiEncoderEmbedding": ...

    def to(self, device) -> "BiEncoderEmbedding":
        if isinstance(device, BiEncoderEmbedding):
            device = device.device
            self.embeddings.to()
        self.embeddings = self.embeddings.to(device)
        self.scoring_mask = self.scoring_mask.to(device)
        return self

    @property
    def device(self) -> torch.device:
        if self.embeddings.device != self.scoring_mask.device:
            raise ValueError("Embeddings and scoring_mask must be on the same device")
        return self.embeddings.device

    def items(self) -> Iterable[Tuple[str, torch.Tensor]]:
        for field in self.__dataclass_fields__:
            yield field, getattr(self, field)


@dataclass
class BiEncoderOutput(LightningIROutput):
    query_embeddings: BiEncoderEmbedding | None = None
    doc_embeddings: BiEncoderEmbedding | None = None


class BiEncoderModel(LightningIRModel):

    _tied_weights_keys = ["projection.decoder.bias", "projection.decoder.weight", "encoder.embed_tokens.weight"]
    _keys_to_ignore_on_load_unexpected = [r"decoder"]

    config_class = BiEncoderConfig

    def __init__(self, config: BiEncoderConfig, *args, **kwargs) -> None:
        super().__init__(config, *args, **kwargs)
        self.config: BiEncoderConfig
        self.scoring_function = ScoringFunction(self.config)
        self.projection: torch.nn.Linear | MLMHead | None = None
        if self.config.projection is not None:
            if "linear" in self.config.projection:
                self.projection = torch.nn.Linear(
                    self.config.hidden_size,
                    self.config.embedding_dim,
                    bias="no_bias" not in self.config.projection,
                )
            elif self.config.projection == "mlm":
                self.projection = MLMHead(config)
            else:
                raise ValueError(f"Unknown projection {self.config.projection}")
        else:
            if self.config.embedding_dim != self.config.hidden_size:
                warnings.warn(
                    "No projection is used but embedding_dim != hidden_size. "
                    "The output embeddings will not have embedding_size dimensions."
                )

        self.query_mask_scoring_input_ids: torch.Tensor | None = None
        self.doc_mask_scoring_input_ids: torch.Tensor | None = None
        self.add_mask_scoring_input_ids()

    @classmethod
    def _load_pretrained_model(
        cls, model, state_dict, loaded_keys, resolved_archive_file, pretrained_model_name_or_path, *args, **kwargs
    ):
        if model.config.projection == "mlm":
            has_base_model_prefix = any(s.startswith(model.base_model_prefix) for s in state_dict.keys())
            prefix = model.base_model_prefix + "." if has_base_model_prefix else ""
            for key in list(state_dict.keys()):
                if key.startswith("cls"):
                    new_key = prefix + key.replace("cls.predictions", "projection").replace(".transform", "")
                    state_dict[new_key] = state_dict.pop(key)
                    loaded_keys[loaded_keys.index(key)] = new_key
        return super()._load_pretrained_model(
            model, state_dict, loaded_keys, resolved_archive_file, pretrained_model_name_or_path, *args, **kwargs
        )

    def get_output_embeddings(self) -> torch.nn.Module | None:
        if isinstance(self.projection, MLMHead):
            return self.projection.decoder
        return None

    def add_mask_scoring_input_ids(self) -> None:
        for sequence in ("query", "doc"):
            mask_scoring_tokens = getattr(self.config, f"{sequence}_mask_scoring_tokens")
            if mask_scoring_tokens is None:
                continue
            if mask_scoring_tokens == "punctuation":
                mask_scoring_tokens = list(punctuation)
            try:
                from transformers import AutoTokenizer

                tokenizer = AutoTokenizer.from_pretrained(self.config.name_or_path)
            except OSError:
                raise ValueError("Can't use token scoring masking if the checkpoint does not have a tokenizer.")
            mask_scoring_input_ids = []
            for token in mask_scoring_tokens:
                if token not in tokenizer.vocab:
                    raise ValueError(f"Token {token} not in tokenizer vocab")
                mask_scoring_input_ids.append(tokenizer.vocab[token])
            setattr(
                self,
                f"{sequence}_mask_scoring_input_ids",
                torch.tensor(mask_scoring_input_ids, dtype=torch.long),
            )

    def forward(
        self,
        query_encoding: BatchEncoding | None,
        doc_encoding: BatchEncoding | None,
        num_docs: Sequence[int] | int | None = None,
    ) -> BiEncoderOutput:
        query_embeddings = None
        if query_encoding is not None:
            query_embeddings = self.encode_query(query_encoding)
        doc_embeddings = None
        if doc_encoding is not None:
            doc_embeddings = self.encode_doc(doc_encoding)
        scores = None
        if doc_embeddings is not None and query_embeddings is not None:
            scores = self.score(query_embeddings, doc_embeddings, num_docs)
        return BiEncoderOutput(scores=scores, query_embeddings=query_embeddings, doc_embeddings=doc_embeddings)

    def encode_query(self, encoding: BatchEncoding) -> BiEncoderEmbedding:
        return self._encode(
            encoding,
            expansion=self.config.query_expansion,
            pooling_strategy=self.config.query_pooling_strategy,
            mask_scoring_input_ids=self.query_mask_scoring_input_ids,
        )

    def encode_doc(self, encoding: BatchEncoding) -> BiEncoderEmbedding:
        return self._encode(
            encoding,
            expansion=self.config.doc_expansion,
            pooling_strategy=self.config.doc_pooling_strategy,
            mask_scoring_input_ids=self.doc_mask_scoring_input_ids,
        )

    @_batch_encoding
    def _encode(
        self,
        encoding: BatchEncoding,
        expansion: bool = False,
        pooling_strategy: Literal["first", "mean", "max", "sum"] | None = None,
        mask_scoring_input_ids: torch.Tensor | None = None,
    ) -> BiEncoderEmbedding:
        embeddings = self._backbone_forward(**encoding).last_hidden_state
        if self.projection is not None:
            embeddings = self.projection(embeddings)
        embeddings = self._sparsification(embeddings, self.config.sparsification)
        embeddings = self._pooling(embeddings, encoding["attention_mask"], pooling_strategy)
        if self.config.normalize:
            embeddings = torch.nn.functional.normalize(embeddings, dim=-1)
        scoring_mask = self._scoring_mask(
            encoding["input_ids"],
            encoding["attention_mask"],
            expansion,
            pooling_strategy,
            mask_scoring_input_ids,
        )
        return BiEncoderEmbedding(embeddings, scoring_mask)

    def query_scoring_mask(self, input_ids: torch.Tensor | None, attention_mask: torch.Tensor | None) -> torch.Tensor:
        return self._scoring_mask(
            input_ids,
            attention_mask,
            expansion=self.config.query_expansion,
            pooling_strategy=self.config.query_pooling_strategy,
            mask_scoring_input_ids=self.config.query_mask_scoring_input_ids,
        )

    def doc_scoring_mask(self, input_ids: torch.Tensor | None, attention_mask: torch.Tensor | None) -> torch.Tensor:
        return self._scoring_mask(
            input_ids,
            attention_mask,
            expansion=self.config.query_expansion,
            pooling_strategy=self.config.doc_pooling_strategy,
            mask_scoring_input_ids=self.config.doc_mask_scoring_input_ids,
        )

    def _scoring_mask(
        self,
        input_ids: torch.Tensor | None,
        attention_mask: torch.Tensor | None,
        expansion: bool,
        pooling_strategy: Literal["first", "mean", "max", "sum"] | None = None,
        mask_scoring_input_ids: torch.Tensor | None = None,
    ) -> torch.Tensor:
        if input_ids is not None:
            shape = input_ids.shape
            device = input_ids.device
        elif attention_mask is not None:
            shape = attention_mask.shape
            device = attention_mask.device
        else:
            raise ValueError("Pass either input_ids or attention_mask")
        if pooling_strategy is not None:
            return torch.ones((shape[0], 1), dtype=torch.bool, device=device)
        scoring_mask = attention_mask
        if expansion or scoring_mask is None:
            scoring_mask = torch.ones(shape, dtype=torch.bool, device=device)
        scoring_mask = scoring_mask.bool()
        if mask_scoring_input_ids is not None and input_ids is not None:
            ignore_mask = input_ids[..., None].eq(mask_scoring_input_ids.to(device)).any(-1)
            scoring_mask = scoring_mask & ~ignore_mask
        return scoring_mask

    def score(
        self,
        query_embeddings: BiEncoderEmbedding,
        doc_embeddings: BiEncoderEmbedding,
        num_docs: Sequence[int] | int | None = None,
    ) -> torch.Tensor:
        scores = self.scoring_function.score(query_embeddings, doc_embeddings, num_docs=num_docs)
        return scores


def _batch_scoring(
    similarity_function: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    BATCH_SIZE = 1024

    @wraps(similarity_function)
    def batch_similarity_function(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        if x.shape[0] <= BATCH_SIZE:
            return similarity_function(x, y)
        out = torch.zeros(x.shape[0], x.shape[1], y.shape[2], device=x.device, dtype=x.dtype)
        for i in range(0, x.shape[0], BATCH_SIZE):
            out[i : i + BATCH_SIZE] = similarity_function(x[i : i + BATCH_SIZE], y[i : i + BATCH_SIZE])
        return out

    return batch_similarity_function


class ScoringFunction(torch.nn.Module):
    def __init__(self, config: BiEncoderConfig) -> None:
        super().__init__()
        self.config = config
        if self.config.similarity_function == "cosine":
            self.similarity_function = self.cosine_similarity
        elif self.config.similarity_function == "l2":
            self.similarity_function = self.l2_similarity
        elif self.config.similarity_function == "dot":
            self.similarity_function = self.dot_similarity
        else:
            raise ValueError(f"Unknown similarity function {self.config.similarity_function}")
        self.query_aggregation_function = self.config.query_aggregation_function

    def compute_similarity(
        self, query_embeddings: BiEncoderEmbedding, doc_embeddings: BiEncoderEmbedding
    ) -> torch.Tensor:
        # if torch.cuda.is_available():
        #     # bfloat16 similarity yields weird values with gpu, so we use fp16 instead
        #     # TODO investigate why, all values are a factor of 1/4
        #     query_tensor = query_tensor.cuda().half()
        #     doc_tensor = doc_tensor.cuda().half()

        # TODO compute similarity only for non-masked values
        similarity = self.similarity_function(query_embeddings.embeddings, doc_embeddings.embeddings)
        return similarity

    @staticmethod
    @_batch_scoring
    def cosine_similarity(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.cosine_similarity(x, y, dim=-1)

    @staticmethod
    @_batch_scoring
    def l2_similarity(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return -1 * torch.cdist(x, y).squeeze(-2)

    @staticmethod
    @_batch_scoring
    def dot_similarity(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.matmul(x, y.transpose(-1, -2)).squeeze(-2)

    def parse_num_docs(
        self,
        query_embeddings: BiEncoderEmbedding,
        doc_embeddings: BiEncoderEmbedding,
        num_docs: int | Sequence[int] | None,
    ) -> torch.Tensor:
        batch_size = query_embeddings.embeddings.shape[0]
        if isinstance(num_docs, int):
            num_docs = [num_docs] * batch_size
        if isinstance(num_docs, list):
            if sum(num_docs) != doc_embeddings.embeddings.shape[0] or len(num_docs) != batch_size:
                raise ValueError("Num docs does not match doc embeddings")
        if num_docs is None:
            if doc_embeddings.embeddings.shape[0] % batch_size != 0:
                raise ValueError("Docs are not evenly distributed in _batch, but no num_docs provided")
            num_docs = [doc_embeddings.embeddings.shape[0] // batch_size] * batch_size
        return torch.tensor(num_docs, device=query_embeddings.embeddings.device)

    def expand_query_embeddings(
        self,
        embeddings: BiEncoderEmbedding,
        num_docs: torch.Tensor,
    ) -> BiEncoderEmbedding:
        return BiEncoderEmbedding(
            embeddings.embeddings.repeat_interleave(num_docs, dim=0).unsqueeze(2),
            embeddings.scoring_mask.repeat_interleave(num_docs, dim=0).unsqueeze(2),
        )

    def expand_doc_embeddings(
        self,
        embeddings: BiEncoderEmbedding,
        num_docs: torch.Tensor,
    ) -> BiEncoderEmbedding:
        return BiEncoderEmbedding(embeddings.embeddings.unsqueeze(1), embeddings.scoring_mask.unsqueeze(1))

    def aggregate(
        self,
        scores: torch.Tensor,
        mask: torch.Tensor | None,
        query_aggregation_function: Literal["max", "sum", "mean", "harmonic_mean"] | None,
        dim: int,
    ) -> torch.Tensor:
        if query_aggregation_function is None:
            return scores
        if query_aggregation_function == "max":
            if mask is not None:
                scores = scores.masked_fill(~mask, float("-inf"))
            return scores.max(dim, keepdim=True).values
        if query_aggregation_function == "sum":
            if mask is not None:
                scores = scores.masked_fill(~mask, 0)
            return scores.sum(dim, keepdim=True)
        if mask is None:
            shape = list(scores.shape)
            shape[dim] = 1
            num_non_masked = torch.full(shape, scores.shape[dim], device=scores.device)
        else:
            num_non_masked = mask.sum(dim, keepdim=True)
        if query_aggregation_function == "mean":
            return torch.where(num_non_masked == 0, 0, scores.sum(dim, keepdim=True) / num_non_masked)
        if query_aggregation_function == "harmonic_mean":
            return torch.where(
                num_non_masked == 0,
                0,
                num_non_masked / (1 / scores).sum(dim, keepdim=True),
            )
        raise ValueError(f"Unknown aggregation {query_aggregation_function}")

    def score(
        self,
        query_embeddings: BiEncoderEmbedding,
        doc_embeddings: BiEncoderEmbedding,
        num_docs: Sequence[int] | int | None = None,
    ) -> torch.Tensor:
        num_docs_t = self.parse_num_docs(query_embeddings, doc_embeddings, num_docs)
        query_embeddings = self.expand_query_embeddings(query_embeddings, num_docs_t)
        doc_embeddings = self.expand_doc_embeddings(doc_embeddings, num_docs_t)
        similarity = self.compute_similarity(query_embeddings, doc_embeddings)
        scores = self.aggregate(similarity, doc_embeddings.scoring_mask, "max", -1)
        scores = self.aggregate(scores, query_embeddings.scoring_mask, self.query_aggregation_function, -2)
        return scores[..., 0, 0]



import json
import os
from os import PathLike
from typing import Any, Dict, Literal, Sequence, Tuple

from ..base import LightningIRConfig


class BiEncoderConfig(LightningIRConfig):
    """The configuration class to instantiate a Bi-Encoder model."""

    model_type = "bi-encoder"

    TOKENIZER_ARGS = LightningIRConfig.TOKENIZER_ARGS.union(
        {
            "query_expansion",
            "attend_to_query_expanded_tokens",
            "doc_expansion",
            "attend_to_doc_expanded_tokens",
            "add_marker_tokens",
        }
    )

    ADDED_ARGS = LightningIRConfig.ADDED_ARGS.union(
        {
            "similarity_function",
            "query_pooling_strategy",
            "query_mask_scoring_tokens",
            "query_aggregation_function",
            "doc_pooling_strategy",
            "doc_mask_scoring_tokens",
            "normalize",
            "sparsification",
            "embedding_dim",
            "projection",
        }
    ).union(TOKENIZER_ARGS)

    def __init__(
        self,
        similarity_function: Literal["cosine", "dot"] = "dot",
        query_expansion: bool = False,
        attend_to_query_expanded_tokens: bool = False,
        query_pooling_strategy: Literal["first", "mean", "max", "sum"] | None = "mean",
        query_mask_scoring_tokens: Sequence[str] | Literal["punctuation"] | None = None,
        query_aggregation_function: Literal["sum", "mean", "max", "harmonic_mean"] = "sum",
        doc_expansion: bool = False,
        attend_to_doc_expanded_tokens: bool = False,
        doc_pooling_strategy: Literal["first", "mean", "max", "sum"] | None = "mean",
        doc_mask_scoring_tokens: Sequence[str] | Literal["punctuation"] | None = None,
        normalize: bool = False,
        sparsification: Literal["relu", "relu_log"] | None = None,
        add_marker_tokens: bool = False,
        embedding_dim: int = 768,
        projection: Literal["linear", "linear_no_bias", "mlm"] | None = "linear",
        **kwargs,
    ):
        """Initializes a bi-encoder configuration.

        :param similarity_function: Similarity function to compute scores between query and document embeddings,
            defaults to "dot"
        :type similarity_function: Literal['cosine', 'dot'], optional
        :param query_expansion: Whether to expand queries with mask tokens, defaults to False
        :type query_expansion: bool, optional
        :param attend_to_query_expanded_tokens: Whether to allow query tokens to attend to mask tokens,
            defaults to False
        :type attend_to_query_expanded_tokens: bool, optional
        :param query_pooling_strategy: Whether and how to pool the query token embeddings, defaults to "mean"
        :type query_pooling_strategy: Literal['first', 'mean', 'max', 'sum'] | None, optional
        :param query_mask_scoring_tokens: Whether and which query tokens to ignore during scoring, defaults to None
        :type query_mask_scoring_tokens: Sequence[str] | Literal['punctuation'] | None, optional
        :param query_aggregation_function: How to aggregate similarity scores over query tokens, defaults to "sum"
        :type query_aggregation_function: Literal[ 'sum', 'mean', 'max', 'harmonic_mean' ], optional
        :param doc_expansion: Whether to expand documents with mask tokens, defaults to False
        :type doc_expansion: bool, optional
        :param attend_to_doc_expanded_tokens: Whether to allow document tokens to attend to mask tokens,
            defaults to False
        :type attend_to_doc_expanded_tokens: bool, optional
        :param doc_pooling_strategy: Whether andhow to pool document token embeddings, defaults to "mean"
        :type doc_pooling_strategy: Literal['first', 'mean', 'max', 'sum'] | None, optional
        :param doc_mask_scoring_tokens: Whether and which document tokens to ignore during scoring, defaults to None
        :type doc_mask_scoring_tokens: Sequence[str] | Literal['punctuation'] | None, optional
        :param normalize: Whether to normalize query and document embeddings, defaults to False
        :type normalize: bool, optional
        :param sparsification: Whether and which sparsification function to apply, defaults to None
        :type sparsification: Literal['relu', 'relu_log'] | None, optional
        :param add_marker_tokens: Whether to add extra marker tokens [Q] / [D] to queries / documents, defaults to False
        :type add_marker_tokens: bool, optional
        :param embedding_dim: The output embedding dimension, defaults to 768
        :type embedding_dim: int, optional
        :param projection: Whether and how to project the output emeddings, defaults to "linear"
        :type projection: Literal['linear', 'linear_no_bias', 'mlm'] | None, optional
        """
        super().__init__(**kwargs)
        self.similarity_function = similarity_function
        self.query_expansion = query_expansion
        self.attend_to_query_expanded_tokens = attend_to_query_expanded_tokens
        self.query_pooling_strategy = query_pooling_strategy
        self.query_mask_scoring_tokens = query_mask_scoring_tokens
        self.query_aggregation_function = query_aggregation_function
        self.doc_expansion = doc_expansion
        self.attend_to_doc_expanded_tokens = attend_to_doc_expanded_tokens
        self.doc_pooling_strategy = doc_pooling_strategy
        self.doc_mask_scoring_tokens = doc_mask_scoring_tokens
        self.normalize = normalize
        self.sparsification = sparsification
        self.add_marker_tokens = add_marker_tokens
        self.embedding_dim = embedding_dim
        self.projection = projection

    def to_dict(self) -> Dict[str, Any]:
        output = super().to_dict()
        if "query_mask_scoring_tokens" in output:
            output.pop("query_mask_scoring_tokens")
        if "doc_mask_scoring_tokens" in output:
            output.pop("doc_mask_scoring_tokens")
        return output

    def save_pretrained(self, save_directory: str | PathLike, push_to_hub: bool = False, **kwargs):
        with open(os.path.join(save_directory, "mask_scoring_tokens.json"), "w") as f:
            json.dump({"query": self.query_mask_scoring_tokens, "doc": self.doc_mask_scoring_tokens}, f)
        return super().save_pretrained(save_directory, push_to_hub, **kwargs)

    @classmethod
    def get_config_dict(
        cls, pretrained_model_name_or_path: str | PathLike, **kwargs
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        config_dict, kwargs = super().get_config_dict(pretrained_model_name_or_path, **kwargs)
        mask_scoring_tokens = None
        mask_scoring_tokens_path = os.path.join(pretrained_model_name_or_path, "mask_scoring_tokens.json")
        if os.path.exists(mask_scoring_tokens_path):
            with open(mask_scoring_tokens_path) as f:
                mask_scoring_tokens = json.load(f)
            config_dict["query_mask_scoring_tokens"] = mask_scoring_tokens["query"]
            config_dict["doc_mask_scoring_tokens"] = mask_scoring_tokens["doc"]
        return config_dict, kwargs



import warnings
from collections import defaultdict
from dataclasses import dataclass
from functools import partial, wraps
from pathlib import Path
from typing import Any, Callable, Literal, Mapping, Sequence, Type, TypeVar

import torch
from transformers import MODEL_MAPPING, BatchEncoding, BertModel
from transformers.modeling_outputs import ModelOutput

from ..flash import FLASH_ATTENTION_MAP
from .class_factory import LightningIRModelClassFactory
from .config import LightningIRConfig
from .external_model_hub import CHECKPOINT_MAPPING, POST_LOAD_CALLBACKS, STATE_DICT_KEY_MAPPING


@dataclass
class LightningIROutput(ModelOutput):
    """Base class for the output of the LightningIR model. It is a subclass of transformers.ModelOutput_.

    .. _transformers.ModelOutput: https://huggingface.co/transformers/main_classes/output.html#transformers.ModelOutput

    :param scores: Output relevance scores for query--document pairs, defaults to None
    :type scores: torch.Tensor | None, optional
    """

    scores: torch.Tensor | None = None


class LightningIRModel:
    """Base class for LightningIR models. Derived classes implement the forward method for handling query
    and document embeddings. It acts as mixin for a transformers.PreTrainedModel_ backbone model.

    .. _transformers.PreTrainedModel: \
https://huggingface.co/transformers/main_classes/model.html#transformers.PreTrainedModel
    """

    config_class: Type[LightningIRConfig] = LightningIRConfig
    """Configuration class for the model."""

    ALLOW_SUB_BATCHING = True
    """Flag to allow mini batches of documents for a single query. Set to false for listwise models to ensure
    correctness."""

    def __init__(self, config: LightningIRConfig, *args, **kwargs) -> None:
        """Initializes the model.

        :param config: Configuration class for the model
        :type config: LightningIRConfig
        """
        super().__init__(config, *args, **kwargs)
        self.config = config

        self._sub_batch_size: int | None = None

        if self.config.backbone_model_type is not None:
            flash_attn = FLASH_ATTENTION_MAP.get(self.config.backbone_model_type, None)
            if flash_attn is not None:
                flash_attn_forward, self_attn_pattern = flash_attn
                for name, module in self.named_modules():
                    if name.endswith(self_attn_pattern):
                        module.forward = partial(flash_attn_forward, module)

    def _backbone_forward(self, *args, **kwargs):
        raise NotImplementedError

    def forward(self, *args, **kwargs) -> LightningIROutput:
        """Forward method of the model. Must be implemented by the derived class."""
        raise NotImplementedError

    def _sparsification(
        self, embeddings: torch.Tensor, sparsification_strategy: Literal["relu", "relu_log"] | None = None
    ) -> torch.Tensor:
        """Helper method to apply sparsification to the embeddings.

        :param embeddings: Query or document embeddings
        :type embeddings: torch.Tensor
        :param sparsification_strategy: The sparsification strategy. No sparsification is applied if None,
        defaults to None
        :type sparsification_strategy: Literal[&quot;relu&quot;, &quot;relu_log&quot;] | None, optional
        :raises ValueError: If an unknown sparsification strategy is passed
        :return: (Optionally) sparsified embeddings
        :rtype: torch.Tensor
        """
        if sparsification_strategy is None:
            return embeddings
        if sparsification_strategy == "relu":
            return torch.relu(embeddings)
        if sparsification_strategy == "relu_log":
            return torch.log1p(torch.relu(embeddings))
        raise ValueError(f"Unknown sparsification strategy: {sparsification_strategy}")

    def _pooling(
        self,
        embeddings: torch.Tensor,
        attention_mask: torch.Tensor | None,
        pooling_strategy: Literal["first", "mean", "max", "sum"] | None,
    ) -> torch.Tensor:
        """Helper method to apply pooling to the embeddings.

        :param embeddings: Query or document embeddings
        :type embeddings: torch.Tensor
        :param attention_mask: Query or document attention mask
        :type attention_mask: torch.Tensor | None
        :param pooling_strategy: The pooling strategy. No pooling is applied if None.
        :type pooling_strategy: Literal[&quot;first&quot;, &quot;mean&quot;, &quot;max&quot;, &quot;sum&quot;] | None
        :raises ValueError: If an unknown pooling strategy is passed
        :return: (Optionally) pooled embeddings
        :rtype: torch.Tensor
        """
        if pooling_strategy is None:
            return embeddings
        if pooling_strategy == "first":
            return embeddings[:, [0]]
        if pooling_strategy in ("sum", "mean"):
            if attention_mask is not None:
                embeddings = embeddings * attention_mask.unsqueeze(-1)
            embeddings = embeddings.sum(dim=1, keepdim=True)
            if pooling_strategy == "mean":
                if attention_mask is not None:
                    embeddings = embeddings / attention_mask.sum(dim=1, keepdim=True).unsqueeze(-1)
            return embeddings
        if pooling_strategy == "max":
            if attention_mask is not None:
                embeddings = embeddings.masked_fill(~attention_mask.bool().unsqueeze(-1), -1e9)
            return embeddings.max(dim=1, keepdim=True).values
        raise ValueError(f"Unknown pooling strategy: {self.pooling_strategy}")

    @classmethod
    def _load_pretrained_model(
        cls, model, state_dict, loaded_keys, resolved_archive_file, pretrained_model_name_or_path, *args, **kwargs
    ):
        if pretrained_model_name_or_path in STATE_DICT_KEY_MAPPING:
            map_keys = STATE_DICT_KEY_MAPPING[pretrained_model_name_or_path]
            for orig_key, new_key in map_keys:
                if orig_key is not None:
                    state_dict[new_key] = state_dict.pop(orig_key)
                    loaded_keys[loaded_keys.index(orig_key)] = new_key
                else:
                    loaded_keys.append(new_key)
        model, *out = super()._load_pretrained_model(
            model, state_dict, loaded_keys, resolved_archive_file, pretrained_model_name_or_path, *args, **kwargs
        )
        if pretrained_model_name_or_path in POST_LOAD_CALLBACKS:
            model = POST_LOAD_CALLBACKS[pretrained_model_name_or_path](model)
        return (model, *out)

    @classmethod
    def from_pretrained(cls, model_name_or_path: str | Path, *args, **kwargs) -> "LightningIRModel":
        """Loads a pretrained model. Wraps the transformers.PreTrainedModel.from_pretrained_ method and to return a
        derived LightningIRModel. See :class:`LightningIRModelClassFactory` for more details.

        .. _transformers.PreTrainedModel.from_pretrained: https://huggingface.co/transformers/main_classes/model.html#transformers.PreTrainedModel.from_pretrained # noqa

        :param model_name_or_path: Name or path of the pretrained model
        :type model_name_or_path: str | Path
        :raises ValueError: If called on the abstract class :class:`LightningIRModel` and no config is passed
        :return: A derived LightningIRModel consisting of a backbone model and a LightningIRModel mixin
        :rtype: LightningIRModel

        .. ::doctest
        .. highlight:: python
        .. code-block:: python

            >>> # Loading using model class and backbone checkpoint
            >>> type(CrossEncoderModel.from_pretrained("bert-base-uncased"))
            <class 'lightning_ir.base.class_factory.CrossEncoderBertModel'>
            >>> # Loading using base class and backbone checkpoint
            >>> type(LightningIRModel.from_pretrained("bert-base-uncased", config=CrossEncoderConfig()))
            <class 'lightning_ir.base.class_factory.CrossEncoderBertModel'>
        """
        # provides AutoModel.from_pretrained support
        config = kwargs.get("config", None)
        if cls is LightningIRModel or all(issubclass(base, LightningIRModel) for base in cls.__bases__):
            # no backbone models found, create derived lightning-ir model based on backbone model
            if model_name_or_path in CHECKPOINT_MAPPING:
                _config = CHECKPOINT_MAPPING[model_name_or_path]
                config_class = _config.__class__
                if config is not None:
                    warnings.warn(f"{model_name_or_path} is a registered checkpoint. The provided config is ignored.")
                config = _config
            elif config is not None:
                config_class = config.__class__
            elif cls is not LightningIRModel:
                config_class = cls.config_class
            else:
                config_class = LightningIRModelClassFactory.get_lightning_ir_config(model_name_or_path)
                if config_class is None:
                    raise ValueError("Pass a config to `from_pretrained`.")
            BackboneConfig = LightningIRModelClassFactory.get_backbone_config(model_name_or_path)
            BackboneModel = MODEL_MAPPING[BackboneConfig]
            cls = LightningIRModelClassFactory(config_class).from_backbone_class(BackboneModel)
            if config is not None and all(issubclass(base, LightningIRConfig) for base in config.__class__.__bases__):
                derived_config = cls.config_class.from_pretrained(model_name_or_path, config=config)
                derived_config.update(config.to_dict())
                kwargs["config"] = derived_config
            return cls.from_pretrained(model_name_or_path, *args, **kwargs)
        if issubclass(cls, BertModel):
            kwargs["add_pooling_layer"] = False
        return super(LightningIRModel, cls).from_pretrained(model_name_or_path, *args, **kwargs)


T = TypeVar("T")


def _cat_outputs(
    outputs: Sequence[Mapping] | Sequence[torch.Tensor] | Sequence[None], OutputClass: Type[T] | None
) -> torch.Tensor | T | None:
    if len(outputs) == 1:
        return outputs[0]
    if len(outputs) == 0 or outputs[0] is None or OutputClass is None:
        return None
    if isinstance(outputs[0], torch.Tensor):
        return torch.cat(outputs, dim=0)
    agg = defaultdict(list)
    types = {}
    for output in outputs:
        for key, value in output.items():
            agg[key].append(value)
            types[key] = type(value)
    return OutputClass(**{key: _cat_outputs(value, types[key]) for key, value in agg.items()})


def _batch_encoding(
    func: Callable[[LightningIRModel, BatchEncoding, ...], Any]
) -> Callable[[LightningIRModel, BatchEncoding, ...], Any]:

    @wraps(func)
    def wrapper(self, encoding: BatchEncoding, *args, **kwargs) -> Any:
        if not self.ALLOW_SUB_BATCHING:
            return func(self, encoding, *args, **kwargs)
        sub_batch_size = self._sub_batch_size or encoding.input_ids.shape[0]
        sub_encoding = encoding
        remaining_encoding = encoding
        OutputClass = None
        outputs = []
        while True:
            try:
                # ceil division
                num_batches = -(remaining_encoding.input_ids.shape[0] // -sub_batch_size)
                for _ in range(num_batches):
                    sub_encoding = BatchEncoding(
                        {key: value[:sub_batch_size] for key, value in remaining_encoding.items()}
                    )
                    output = func(self, sub_encoding, *args, **kwargs)
                    OutputClass = output.__class__
                    outputs.append(output)
                    remaining_encoding = BatchEncoding(
                        {key: value[sub_batch_size:] for key, value in remaining_encoding.items()}
                    )
                break
            except RuntimeError as e:
                if "CUDA out of memory" in str(e) or "CUDACachingAllocator.cpp" in str(e):
                    self._sub_batch_size = sub_batch_size = sub_batch_size // 2
                    if sub_batch_size == 0:
                        raise e
                else:
                    raise e
        if OutputClass is None:
            raise ValueError("No output was generated.")
        return _cat_outputs(outputs, OutputClass)

    return wrapper



from typing import Literal

from ..base import LightningIRConfig


class CrossEncoderConfig(LightningIRConfig):
    model_type = "cross-encoder"

    ADDED_ARGS = LightningIRConfig.ADDED_ARGS.union({"pooling_strategy", "linear_bias"})

    def __init__(
        self,
        query_length: int = 32,
        doc_length: int = 512,
        pooling_strategy: Literal["first", "mean", "max", "sum"] = "first",
        linear_bias: bool = False,
        **kwargs
    ):
        super().__init__(query_length=query_length, doc_length=doc_length, **kwargs)
        self.pooling_strategy = pooling_strategy
        self.linear_bias = linear_bias



import warnings
from abc import abstractmethod
from pathlib import Path

import torch

from ..bi_encoder import BiEncoderConfig, BiEncoderOutput
from ..data import IndexBatch
from .indexer import IndexConfig, Indexer


class FaissIndexer(Indexer):
    INDEX_FACTORY: str

    def __init__(
        self,
        index_dir: Path,
        index_config: "FaissIndexConfig",
        bi_encoder_config: BiEncoderConfig,
        verbose: bool = False,
    ) -> None:
        super().__init__(index_dir, index_config, bi_encoder_config, verbose)
        import faiss

        similarity_function = bi_encoder_config.similarity_function
        if similarity_function in ("cosine", "dot"):
            self.metric_type = faiss.METRIC_INNER_PRODUCT
        else:
            raise ValueError(f"similarity_function {similarity_function} unknown")

        index_factory = self.INDEX_FACTORY.format(**index_config.to_dict())
        if similarity_function == "cosine":
            index_factory = "L2norm," + index_factory
        self.index = faiss.index_factory(self.bi_encoder_config.embedding_dim, index_factory, self.metric_type)

        self.set_verbosity()

        if torch.cuda.is_available():
            self.to_gpu()

    @abstractmethod
    def to_gpu(self) -> None: ...

    @abstractmethod
    def to_cpu(self) -> None: ...

    @abstractmethod
    def set_verbosity(self, verbose: bool | None = None) -> None: ...

    def process_embeddings(self, embeddings: torch.Tensor) -> torch.Tensor:
        return embeddings

    def save(self) -> None:
        super().save()
        import faiss

        if self.num_embeddings != self.index.ntotal:
            raise ValueError("number of embeddings does not match index.ntotal")
        if torch.cuda.is_available():
            self.index = faiss.index_gpu_to_cpu(self.index)

        faiss.write_index(self.index, str(self.index_dir / "index.faiss"))

    def add(self, index_batch: IndexBatch, output: BiEncoderOutput) -> None:
        doc_embeddings = output.doc_embeddings
        if doc_embeddings is None:
            raise ValueError("Expected doc_embeddings in BiEncoderOutput")
        doc_lengths = doc_embeddings.scoring_mask.sum(dim=1)
        embeddings = doc_embeddings.embeddings[doc_embeddings.scoring_mask]
        doc_ids = index_batch.doc_ids
        embeddings = self.process_embeddings(embeddings)

        if embeddings.shape[0]:
            self.index.add(embeddings.float().cpu())

        self.num_embeddings += embeddings.shape[0]
        self.num_docs += len(doc_ids)

        self.doc_lengths.extend(doc_lengths.cpu().tolist())
        self.doc_ids.extend(doc_ids)


class FaissFlatIndexer(FaissIndexer):
    INDEX_FACTORY = "Flat"

    def __init__(
        self,
        index_dir: Path,
        index_config: "FaissFlatIndexConfig",
        bi_encoder_config: BiEncoderConfig,
        verbose: bool = False,
    ) -> None:
        super().__init__(index_dir, index_config, bi_encoder_config, verbose)
        self.index_config: FaissFlatIndexConfig

    def to_gpu(self) -> None:
        pass

    def to_cpu(self) -> None:
        pass

    def set_verbosity(self, verbose: bool | None = None) -> None:
        self.index.verbose = self.verbose if verbose is None else verbose


class FaissIVFIndexer(FaissIndexer):
    INDEX_FACTORY = "IVF{num_centroids},Flat"

    def __init__(
        self,
        index_dir: Path,
        index_config: "FaissIVFIndexConfig",
        bi_encoder_config: BiEncoderConfig,
        verbose: bool = False,
    ) -> None:
        super().__init__(index_dir, index_config, bi_encoder_config, verbose)

        import faiss

        ivf_index = faiss.extract_index_ivf(self.index)
        if hasattr(ivf_index, "quantizer"):
            quantizer = ivf_index.quantizer
            if hasattr(faiss.downcast_index(quantizer), "hnsw"):
                downcasted_quantizer = faiss.downcast_index(quantizer)
                downcasted_quantizer.hnsw.efConstruction = index_config.ef_construction

        # default faiss values
        # https://github.com/facebookresearch/faiss/blob/dafdff110489db7587b169a0afee8470f220d295/faiss/Clustering.h#L43
        max_points_per_centroid = 256
        self.num_train_embeddings = (
            index_config.num_train_embeddings or index_config.num_centroids * max_points_per_centroid
        )

        self._train_embeddings = torch.full(
            (
                self.num_train_embeddings,
                self.bi_encoder_config.embedding_dim,
            ),
            torch.nan,
            dtype=torch.float32,
        )

    def to_gpu(self) -> None:
        import faiss

        # clustering_index overrides the index used during clustering but leaves
        # the quantizer on the gpu
        # https://faiss.ai/cpp_api/namespace/namespacefaiss_1_1gpu.html
        clustering_index = faiss.index_cpu_to_all_gpus(
            faiss.IndexFlat(self.bi_encoder_config.embedding_dim, self.metric_type)
        )
        clustering_index.verbose = self.verbose
        index_ivf = faiss.extract_index_ivf(self.index)
        index_ivf.clustering_index = clustering_index

    def to_cpu(self) -> None:
        import faiss

        self.index = faiss.index_gpu_to_cpu(self.index)

        # https://gist.github.com/mdouze/334ad6a979ac3637f6d95e9091356d3e
        # move index to cpu but leave quantizer on gpu
        index_ivf = faiss.extract_index_ivf(self.index)
        quantizer = index_ivf.quantizer
        gpu_quantizer = faiss.index_cpu_to_gpu(faiss.StandardGpuResources(), 0, quantizer)
        index_ivf.quantizer = gpu_quantizer

    def process_embeddings(self, embeddings: torch.Tensor) -> torch.Tensor:
        embeddings = self._grab_train_embeddings(embeddings)
        self._train()
        return embeddings

    def _grab_train_embeddings(self, embeddings: torch.Tensor) -> torch.Tensor:
        if self._train_embeddings is not None:
            # save training embeddings until num_train_embeddings is reached
            # if num_train_embeddings overflows, save the remaining embeddings
            start = self.num_embeddings
            end = start + embeddings.shape[0]
            if end > self.num_train_embeddings:
                end = self.num_train_embeddings
            length = end - start
            self._train_embeddings[start:end] = embeddings[:length]
            self.num_embeddings += length
            embeddings = embeddings[length:]
        return embeddings

    def _train(self, force: bool = False):
        if self._train_embeddings is not None and (force or self.num_embeddings >= self.num_train_embeddings):
            if torch.isnan(self._train_embeddings).any():
                warnings.warn(
                    "Corpus contains less tokens/documents than num_train_embeddings. " "Removing NaN embeddings."
                )
                self._train_embeddings = self._train_embeddings[~torch.isnan(self._train_embeddings).any(dim=1)]
            self.index.train(self._train_embeddings)
            if torch.cuda.is_available():
                self.to_cpu()
            self.index.add(self._train_embeddings)
            self._train_embeddings = None
            self.set_verbosity(False)

    def save(self) -> None:
        if not self.index.is_trained:
            self._train(force=True)
        return super().save()

    def set_verbosity(self, verbose: bool | None = None) -> None:
        import faiss

        verbose = verbose if verbose is not None else self.verbose
        index = faiss.extract_index_ivf(self.index)
        for elem in (index, index.quantizer):
            setattr(elem, "verbose", verbose)


class FaissIVFPQIndexer(FaissIVFIndexer):
    INDEX_FACTORY = "OPQ{num_subquantizers},IVF{num_centroids}_HNSW32,PQ{num_subquantizers}x{n_bits}"

    def __init__(
        self,
        index_dir: Path,
        index_config: "FaissIVFPQIndexConfig",
        bi_encoder_config: BiEncoderConfig,
        verbose: bool = False,
    ) -> None:
        import faiss

        super().__init__(index_dir, index_config, bi_encoder_config, verbose)
        self.index_config: FaissIVFPQIndexConfig

        index_ivf = faiss.extract_index_ivf(self.index)
        index_ivf.make_direct_map()

    def set_verbosity(self, verbose: bool | None = None) -> None:
        super().set_verbosity(verbose)
        import faiss

        verbose = verbose if verbose is not None else self.verbose
        index_ivf_pq = faiss.downcast_index(self.index.index)
        for elem in (
            index_ivf_pq.pq,
            index_ivf_pq.quantizer,
        ):
            setattr(elem, "verbose", verbose)


class FaissIndexConfig(IndexConfig):
    indexer_class = FaissIndexer

    def to_dict(self) -> dict:
        return self.__dict__.copy()


class FaissFlatIndexConfig(FaissIndexConfig):
    indexer_class = FaissFlatIndexer


class FaissIVFIndexConfig(FaissIndexConfig):
    indexer_class = FaissIVFIndexer

    def __init__(
        self,
        num_train_embeddings: int | None = None,
        num_centroids: int = 262144,
        ef_construction: int = 40,
    ) -> None:
        super().__init__()
        self.num_train_embeddings = num_train_embeddings
        self.num_centroids = num_centroids
        self.ef_construction = ef_construction


class FaissIVFPQIndexConfig(FaissIVFIndexConfig):
    indexer_class = FaissIVFPQIndexer

    def __init__(
        self,
        num_train_embeddings: int | None = None,
        num_centroids: int = 262144,
        ef_construction: int = 40,
        num_subquantizers: int = 16,
        n_bits: int = 8,
    ) -> None:
        super().__init__(num_train_embeddings, num_centroids, ef_construction)
        self.num_subquantizers = num_subquantizers
        self.n_bits = n_bits



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Tuple, Type

from transformers import (
    CONFIG_MAPPING,
    MODEL_MAPPING,
    TOKENIZER_MAPPING,
    AutoConfig,
    AutoModel,
    AutoTokenizer,
    PretrainedConfig,
    PreTrainedModel,
    PreTrainedTokenizerBase,
)
from transformers.models.auto.tokenization_auto import get_tokenizer_config, tokenizer_class_from_name

if TYPE_CHECKING:
    from . import LightningIRConfig, LightningIRModel, LightningIRTokenizer


class LightningIRClassFactory(ABC):
    """Base class for creating derived LightningIR classes from HuggingFace classes."""

    def __init__(self, MixinConfig: Type[LightningIRConfig]) -> None:
        """Creates a new LightningIRClassFactory.

        :param MixinConfig: LightningIRConfig mixin class
        :type MixinConfig: Type[LightningIRConfig]
        """
        if getattr(MixinConfig, "backbone_model_type", None) is not None:
            MixinConfig = MixinConfig.__bases__[0]
        self.MixinConfig = MixinConfig

    @staticmethod
    def get_backbone_config(model_name_or_path: str | Path) -> Type[PretrainedConfig]:
        """Grabs the configuration class from a checkpoint of a pretrained HuggingFace model.

        :param model_name_or_path: Path to the model or its name
        :type model_name_or_path: str | Path
        :return: Configuration class of the backbone model
        :rtype: PretrainedConfig
        """
        backbone_model_type = LightningIRClassFactory.get_backbone_model_type(model_name_or_path)
        return CONFIG_MAPPING[backbone_model_type]

    @staticmethod
    def get_lightning_ir_config(model_name_or_path: str | Path) -> Type[LightningIRConfig] | None:
        """Grabs the LightningIR configuration class from a checkpoint of a pretrained Lightning IR model.

        :param model_name_or_path: Path to the model or its name
        :type model_name_or_path: str | Path
        :return: Configuration class of the Lightning IR model
        :rtype: Type[LightningIRConfig]
        """
        model_type = LightningIRClassFactory.get_lightning_ir_model_type(model_name_or_path)
        if model_type is None:
            return None
        return CONFIG_MAPPING[model_type]

    @staticmethod
    def get_backbone_model_type(model_name_or_path: str | Path, *args, **kwargs) -> str:
        """Grabs the model type from a checkpoint of a pretrained HuggingFace model.

        :param model_name_or_path: Path to the model or its name
        :type model_name_or_path: str | Path
        :return: Model type of the backbone model
        :rtype: str
        """
        config_dict, _ = PretrainedConfig.get_config_dict(model_name_or_path, *args, **kwargs)
        backbone_model_type = config_dict.get("backbone_model_type", None) or config_dict.get("model_type")
        return backbone_model_type

    @staticmethod
    def get_lightning_ir_model_type(model_name_or_path: str | Path) -> str | None:
        """Grabs the Lightning IR model type from a checkpoint of a pretrained HuggingFace model.

        :param model_name_or_path: Path to the model or its name
        :type model_name_or_path: str | Path
        :return: Model type of the Lightning IR model
        :rtype: str | None
        """
        config_dict, _ = PretrainedConfig.get_config_dict(model_name_or_path)
        if "backbone_model_type" not in config_dict:
            return None
        return config_dict.get("model_type", None)

    @property
    def cc_lir_model_type(self) -> str:
        """Camel case model type of the LightningIR model."""
        return "".join(s.title() for s in self.MixinConfig.model_type.split("-"))

    @abstractmethod
    def from_pretrained(self, model_name_or_path: str | Path, *args, **kwargs) -> Any:
        """Loads a derived LightningIR class from a pretrained HuggingFace model. Must be implemented by subclasses.

        :param model_name_or_path: Path to the model or its name
        :type model_name_or_path: str | Path
        :return: Derived LightningIR class
        :rtype: Any
        """
        ...

    @abstractmethod
    def from_backbone_class(self, BackboneClass: Type) -> Type:
        """Creates a derived LightningIR class from a backbone HuggingFace class. Must be implemented by subclasses.

        :param BackboneClass: Backbone class
        :type BackboneClass: Type
        :return: Derived LightningIR class
        :rtype: Type
        """
        ...


class LightningIRConfigClassFactory(LightningIRClassFactory):
    """Class factory for creating derived LightningIRConfig classes from HuggingFace configuration classes."""

    def from_pretrained(self, model_name_or_path: str | Path, *args, **kwargs) -> Type[LightningIRConfig]:
        """Loads a derived LightningIRConfig from a pretrained HuggingFace model.

        :param model_name_or_path: Path to the model or its name
        :type model_name_or_path: str | Path
        :return: Derived LightningIRConfig
        :rtype: Type[LightningIRConfig]
        """
        BackboneConfig = self.get_backbone_config(model_name_or_path)
        DerivedLightningIRConfig = self.from_backbone_class(BackboneConfig)
        return DerivedLightningIRConfig

    def from_backbone_class(self, BackboneClass: Type[PretrainedConfig]) -> Type[LightningIRConfig]:
        """Creates a derived LightningIRConfig from a transformers.PretrainedConfig_ backbone configuration class. If
        the backbone configuration class is already a dervied LightningIRConfig, it is returned as is.

        .. _transformers.PretrainedConfig: \
https://huggingface.co/docs/transformers/main_classes/configuration#transformers.PretrainedConfig

        :param BackboneClass: Backbone configuration class
        :type BackboneClass: Type[PretrainedConfig]
        :return: Derived LightningIRConfig
        :rtype: Type[LightningIRConfig]
        """
        if getattr(BackboneClass, "backbone_model_type", None) is not None:
            return BackboneClass
        LightningIRConfigMixin: Type[LightningIRConfig] = CONFIG_MAPPING[self.MixinConfig.model_type]

        DerivedLightningIRConfig = type(
            f"{self.cc_lir_model_type}{BackboneClass.__name__}",
            (LightningIRConfigMixin, BackboneClass),
            {
                "model_type": f"{BackboneClass.model_type}-{self.MixinConfig.model_type}",
                "backbone_model_type": BackboneClass.model_type,
            },
        )

        AutoConfig.register(DerivedLightningIRConfig.model_type, DerivedLightningIRConfig, exist_ok=True)

        return DerivedLightningIRConfig


class LightningIRModelClassFactory(LightningIRClassFactory):
    """Class factory for creating derived LightningIRModel classes from HuggingFace model classes."""

    def from_pretrained(self, model_name_or_path: str | Path, *args, **kwargs) -> Type[LightningIRModel]:
        """Loads a derived LightningIRModel from a pretrained HuggingFace model.

        :param model_name_or_path: Path to the model or its name
        :type model_name_or_path: str | Path
        :return: Derived LightningIRModel
        :rtype: Type[LightningIRModel]
        """
        BackboneConfig = self.get_backbone_config(model_name_or_path)
        BackboneModel = MODEL_MAPPING[BackboneConfig]
        DerivedLightningIRModel = self.from_backbone_class(BackboneModel)
        return DerivedLightningIRModel

    def from_backbone_class(self, BackboneClass: Type[PreTrainedModel]) -> Type[LightningIRModel]:
        """Creates a derived LightningIRModel from a transformers.PreTrainedModel_ backbone model. If the backbone model
          is already a LightningIRModel, it is returned as is.

        .. _transformers.PreTrainedModel: \
https://huggingface.co/transformers/main_classes/model#transformers.PreTrainedModel

        :param BackboneClass: Backbone model
        :type BackboneClass: Type[PreTrainedModel]
        :raises ValueError: If the backbone model is not a valid backbone model.
        :raises ValueError: If the backbone model is not a LightningIRModel and no LightningIRConfig is passed.
        :raises ValueError: If the LightningIRModel mixin is not registered with the Hugging Face model mapping.
        :return: The derived LightningIRModel
        :rtype: Type[LightningIRModel]
        """
        if getattr(BackboneClass.config_class, "backbone_model_type", None) is not None:
            return BackboneClass
        BackboneConfig = BackboneClass.config_class
        if BackboneConfig is None:
            raise ValueError(
                f"Model {BackboneClass} is not a valid backbone model because it is missing a `config_class`."
            )

        LightningIRModelMixin: Type[LightningIRModel] = MODEL_MAPPING[self.MixinConfig]

        DerivedLightningIRConfig = LightningIRConfigClassFactory(self.MixinConfig).from_backbone_class(BackboneConfig)

        DerivedLightningIRModel = type(
            f"{self.cc_lir_model_type}{BackboneClass.__name__}",
            (LightningIRModelMixin, BackboneClass),
            {"config_class": DerivedLightningIRConfig, "_backbone_forward": BackboneClass.forward},
        )

        AutoModel.register(DerivedLightningIRConfig, DerivedLightningIRModel, exist_ok=True)

        return DerivedLightningIRModel


class LightningIRTokenizerClassFactory(LightningIRClassFactory):
    """Class factory for creating derived LightningIRTokenizer classes from HuggingFace tokenizer classes."""

    @staticmethod
    def get_backbone_config(model_name_or_path: str | Path) -> PretrainedConfig:
        """Grabs the tokenizer configuration class from a checkpoint of a pretrained HuggingFace tokenizer.

        :param model_name_or_path: Path to the tokenizer or its name
        :type model_name_or_path: str | Path
        :return: Configuration class of the backbone tokenizer
        :rtype: PretrainedConfig
        """
        backbone_model_type = LightningIRTokenizerClassFactory.get_backbone_model_type(model_name_or_path)
        return CONFIG_MAPPING[backbone_model_type]

    @staticmethod
    def get_backbone_model_type(model_name_or_path: str | Path, *args, **kwargs) -> str:
        """Grabs the model type from a checkpoint of a pretrained HuggingFace tokenizer.

        :param model_name_or_path: Path to the tokenizer or its name
        :type model_name_or_path: str | Path
        :return: Model type of the backbone tokenizer
        :rtype: str
        """
        try:
            return LightningIRClassFactory.get_backbone_model_type(model_name_or_path, *args, **kwargs)
        except OSError:
            # best guess at model type
            config_dict = get_tokenizer_config(model_name_or_path)
            Tokenizer = tokenizer_class_from_name(config_dict["tokenizer_class"])
            for config, tokenizers in TOKENIZER_MAPPING.items():
                if Tokenizer in tokenizers:
                    return getattr(config, "backbone_model_type", None) or getattr(config, "model_type")
            raise ValueError("No backbone model found in the configuration")

    def from_pretrained(
        self, model_name_or_path: str | Path, *args, use_fast: bool = True, **kwargs
    ) -> Type[LightningIRTokenizer]:
        """Loads a derived LightningIRTokenizer from a pretrained HuggingFace tokenizer.

        :param model_name_or_path: Path to the tokenizer or its name
        :type model_name_or_path: str | Path
        :param use_fast: Whether to use the fast or slow tokenizer, defaults to True
        :type use_fast: bool, optional
        :raises ValueError: If use_fast is True and no fast tokenizer is found
        :raises ValueError: If use_fast is False and no slow tokenizer is found
        :return: Derived LightningIRTokenizer
        :rtype: Type[LightningIRTokenizer]
        """
        BackboneConfig = self.get_backbone_config(model_name_or_path)
        BackboneTokenizers = TOKENIZER_MAPPING[BackboneConfig]
        DerivedLightningIRTokenizers = self.from_backbone_classes(BackboneTokenizers, BackboneConfig)
        if use_fast:
            DerivedLightningIRTokenizer = DerivedLightningIRTokenizers[1]
            if DerivedLightningIRTokenizer is None:
                raise ValueError("No fast tokenizer found.")
        else:
            DerivedLightningIRTokenizer = DerivedLightningIRTokenizers[0]
            if DerivedLightningIRTokenizer is None:
                raise ValueError("No slow tokenizer found.")
        return DerivedLightningIRTokenizer

    def from_backbone_classes(
        self,
        BackboneClasses: Tuple[Type[PreTrainedTokenizerBase] | None, Type[PreTrainedTokenizerBase] | None],
        BackboneConfig: Type[PretrainedConfig] | None = None,
    ) -> Tuple[Type[LightningIRTokenizer] | None, Type[LightningIRTokenizer] | None]:
        """Creates derived slow and fastLightningIRTokenizers from a tuple of backbone HuggingFace tokenizer classes.

        :param BackboneClasses: Slow and fast backbone tokenizer classes
        :type BackboneClasses: Tuple[Type[PreTrainedTokenizerBase] | None, Type[PreTrainedTokenizerBase] | None]
        :param BackboneConfig: Backbone configuration class, defaults to None
        :type BackboneConfig: Type[PretrainedConfig], optional
        :return: Slow and fast derived LightningIRTokenizers
        :rtype: Tuple[Type[LightningIRTokenizer] | None, Type[LightningIRTokenizer] | None]
        """
        DerivedLightningIRTokenizers = tuple(
            None if BackboneClass is None else self.from_backbone_class(BackboneClass)
            for BackboneClass in BackboneClasses
        )
        if DerivedLightningIRTokenizers[1] is not None:
            DerivedLightningIRTokenizers[1].slow_tokenizer_class = DerivedLightningIRTokenizers[0]

        DerivedLightningIRConfig = LightningIRConfigClassFactory(self.MixinConfig).from_backbone_class(BackboneConfig)
        AutoTokenizer.register(
            DerivedLightningIRConfig, DerivedLightningIRTokenizers[0], DerivedLightningIRTokenizers[1]
        )

        return DerivedLightningIRTokenizers

    def from_backbone_class(self, BackboneClass: Type[PreTrainedTokenizerBase]) -> Type[LightningIRTokenizer]:
        """Creates a derived LightningIRTokenizer from a transformers.PreTrainedTokenizerBase_ backbone tokenizer. If
        the backbone tokenizer is already a LightningIRTokenizer, it is returned as is.

        .. _transformers.PreTrainedTokenizerBase: \
https://huggingface.co/transformers/main_classes/tokenizer.html#transformers.PreTrainedTokenizerBase

        :param BackboneClass: Backbone tokenizer class
        :type BackboneClass: Type[PreTrainedTokenizerBase]
        :return: Derived LightningIRTokenizer
        :rtype: Type[LightningIRTokenizer]
        """
        if hasattr(BackboneClass, "config_class"):
            return BackboneClass
        LightningIRTokenizerMixin = TOKENIZER_MAPPING[self.MixinConfig][0]

        DerivedLightningIRTokenizer = type(
            f"{self.cc_lir_model_type}{BackboneClass.__name__}", (LightningIRTokenizerMixin, BackboneClass), {}
        )

        return DerivedLightningIRTokenizer



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Dict, List, Tuple

if TYPE_CHECKING:
    from .config import LightningIRConfig
    from .model import LightningIRModel

CHECKPOINT_MAPPING: Dict[str, LightningIRConfig] = {}
STATE_DICT_KEY_MAPPING: Dict[str, List[Tuple[str | None, str]]] = {}
POST_LOAD_CALLBACKS: Dict[str, Callable[[LightningIRModel], LightningIRModel]] = {}



from typing import Literal

from ..bi_encoder import BiEncoderConfig, BiEncoderModel


class SpladeConfig(BiEncoderConfig):
    model_type = "splade"

    def __init__(
        self,
        query_pooling_strategy: Literal["first", "mean", "max", "sum"] | None = "max",
        doc_pooling_strategy: Literal["first", "mean", "max", "sum"] | None = "max",
        projection: Literal["linear", "linear_no_bias", "mlm"] | None = "mlm",
        sparsification: Literal["relu", "relu_log"] | None = "relu_log",
        embedding_dim: int = 30522,
        **kwargs,
    ) -> None:
        kwargs["query_expansion"] = False
        kwargs["attend_to_query_expanded_tokens"] = False
        kwargs["query_mask_scoring_tokens"] = None
        kwargs["doc_expansion"] = False
        kwargs["attend_to_doc_expanded_tokens"] = False
        kwargs["doc_mask_scoring_tokens"] = None
        kwargs["query_aggregation_function"] = "sum"
        kwargs["normalize"] = False
        kwargs["add_marker_tokens"] = False
        super().__init__(
            query_pooling_strategy=query_pooling_strategy,
            doc_pooling_strategy=doc_pooling_strategy,
            embedding_dim=embedding_dim,
            projection=projection,
            sparsification=sparsification,
            **kwargs,
        )


class SpladeModel(BiEncoderModel):
    config_class = SpladeConfig



from .col import ColConfig, ColModel
from .splade import SpladeConfig, SpladeModel
from .t5 import T5CrossEncoderConfig, T5CrossEncoderModel, T5CrossEncoderTokenizer
from .xtr import XTRConfig, XTRModel

__all__ = [
    "ColConfig",
    "ColModel",
    "SpladeConfig",
    "SpladeModel",
    "T5CrossEncoderConfig",
    "T5CrossEncoderModel",
    "T5CrossEncoderTokenizer",
    "XTRConfig",
    "XTRModel",
]



import codecs
import json
from pathlib import Path
from typing import Any, Dict, Literal, NamedTuple, Tuple, Type

import ir_datasets
from ir_datasets.datasets.base import Dataset
from ir_datasets.formats import BaseDocPairs, jsonl, trec, tsv
from ir_datasets.util import Cache, DownloadConfig, GzipExtract

CONSTITUENT_TYPE_MAP: Dict[str, Dict[str, Type]] = {
    "docs": {
        ".json": jsonl.JsonlDocs,
        ".jsonl": jsonl.JsonlDocs,
        ".tsv": tsv.TsvDocs,
    },
    "queries": {
        ".json": jsonl.JsonlQueries,
        ".jsonl": jsonl.JsonlQueries,
        ".tsv": tsv.TsvQueries,
    },
    "qrels": {".tsv": trec.TrecQrels, ".qrels": trec.TrecQrels},
    "scoreddocs": {".run": trec.TrecScoredDocs, ".tsv": trec.TrecScoredDocs},
    "docpairs": {".tsv": tsv.TsvDocPairs},
}


def load_constituent(
    constituent: str | None,
    constituent_type: Literal["docs", "queries", "qrels", "scoreddocs", "docpairs"],
    **kwargs,
) -> Any:
    if constituent is None:
        return None
    if constituent in ir_datasets.registry._registered:
        return getattr(ir_datasets.load(constituent), f"{constituent_type}_handler")
    constituent_path = Path(constituent)
    if not constituent_path.exists():
        raise ValueError(f"unable to load {constituent}, expected an `ir_datasets` id or valid path")
    suffix = constituent_path.suffixes[0]
    constituent_types = CONSTITUENT_TYPE_MAP[constituent_type]
    if suffix not in constituent_types:
        raise ValueError(f"Unknown file type: {suffix}, expected one of {constituent_types.keys()}")
    ConstituentType = constituent_types[suffix]
    return ConstituentType(Cache(None, constituent_path), **kwargs)


def register_local(
    dataset_id: str,
    docs: str | None = None,
    queries: str | None = None,
    qrels: str | None = None,
    docpairs: str | None = None,
    scoreddocs: str | None = None,
    qrels_defs: Dict[int, str] | None = None,
):
    if dataset_id in ir_datasets.registry._registered:
        return

    docs = load_constituent(docs, "docs")
    queries = load_constituent(queries, "queries")
    qrels = load_constituent(qrels, "qrels", qrels_defs=qrels_defs if qrels_defs is not None else {})
    docpairs = load_constituent(docpairs, "docpairs")
    scoreddocs = load_constituent(scoreddocs, "scoreddocs")

    ir_datasets.registry.register(dataset_id, Dataset(docs, queries, qrels, docpairs, scoreddocs))


class ScoredDocTuple(NamedTuple):
    query_id: str
    doc_ids: Tuple[str, ...]
    scores: Tuple[float, ...] | None
    num_docs: int


class ScoredDocTuples(BaseDocPairs):
    def __init__(self, docpairs_dlc):
        self._docpairs_dlc = docpairs_dlc

    def docpairs_path(self):
        return self._docpairs_dlc.path()

    def docpairs_iter(self):
        file_type = None
        if self._docpairs_dlc.path().suffix == ".json":
            file_type = "json"
        elif self._docpairs_dlc.path().suffix in (".tsv", ".run"):
            file_type = "tsv"
        else:
            raise ValueError(f"Unknown file type: {self._docpairs_dlc.path().suffix}")
        with self._docpairs_dlc.stream() as f:
            f = codecs.getreader("utf8")(f)
            for line in f:
                if file_type == "json":
                    data = json.loads(line)
                    qid, *doc_data = data
                    pids, scores = zip(*doc_data)
                    pids = tuple(str(pid) for pid in pids)
                else:
                    cols = line.rstrip().split()
                    pos_score, neg_score, qid, pid1, pid2 = cols
                    pids = (pid1, pid2)
                    scores = (float(pos_score), float(neg_score))
                yield ScoredDocTuple(str(qid), pids, scores, len(pids))

    def docpairs_cls(self):
        return ScoredDocTuple


def register_kd_docpairs():
    base_id = "msmarco-passage"
    split_id = "train"
    file_id = "kd-docpairs"
    cache_path = "bert_cat_ensemble_msmarcopassage_train_scores_ids.tsv"
    dlc_contents = {
        "url": (
            "https://zenodo.org/record/4068216/files/bert_cat_ensemble_"
            "msmarcopassage_train_scores_ids.tsv?download=1"
        ),
        "expected_md5": "4d99696386f96a7f1631076bcc53ac3c",
        "cache_path": cache_path,
    }
    file_name = f"{split_id}/{file_id}.tsv"
    register_msmarco(base_id, split_id, file_id, cache_path, dlc_contents, file_name, ScoredDocTuples)


def register_colbert_docpairs():
    base_id = "msmarco-passage"
    split_id = "train"
    file_id = "colbert-docpairs"
    cache_path = "colbert_64way.json"
    dlc_contents = {
        "url": (
            "https://huggingface.co/colbert-ir/colbertv2.0_msmarco_64way/" "resolve/main/examples.json?download=true"
        ),
        "expected_md5": "8be0c71e330ac54dcd77fba058d291c7",
        "cache_path": cache_path,
    }
    file_name = f"{split_id}/{file_id}.json"
    register_msmarco(base_id, split_id, file_id, cache_path, dlc_contents, file_name, ScoredDocTuples)


def register_rank_distillm():
    base_id = "msmarco-passage"
    split_id = "train"
    file_id = "rank-distillm/rankzephyr"
    cache_path = "rank-distillm-rankzephyr.run"
    dlc_contents = {
        "url": (
            "https://zenodo.org/records/12528410/files/__rankzephyr-colbert-10000-"
            "sampled-100__msmarco-passage-train-judged.run?download=1"
        ),
        "expected_md5": "49f8dbf2c1ee7a2ca1fe517eda528af6",
        "cache_path": cache_path,
    }
    file_name = f"{split_id}/{file_id}.run"
    register_msmarco(
        base_id,
        split_id,
        file_id,
        cache_path,
        dlc_contents,
        file_name,
        trec.TrecScoredDocs,
    )

    file_id = "rank-distillm/set-encoder"
    cache_path = "rank-distillm-set-encoder.run.gz"
    dlc_contents = {
        "url": (
            "https://zenodo.org/records/12528410/files/__set-encoder-colbert__"
            "msmarco-passage-train-judged.run.gz?download=1"
        ),
        "expected_md5": "1f069d0daa9842a54a858cc660149e1a",
        "cache_path": cache_path,
    }
    file_name = f"{split_id}/{file_id}.run"
    register_msmarco(
        base_id,
        split_id,
        file_id,
        cache_path,
        dlc_contents,
        file_name,
        trec.TrecScoredDocs,
        extract=True,
    )


def register_msmarco(
    base_id: str,
    split_id: str,
    file_id: str,
    cache_path: str,
    dlc_contents: Dict[str, Any],
    file_name: str,
    ConstituentType: Type,
    extract: bool = False,
):
    dataset_id = f"{base_id}/{split_id}/{file_id}"
    if dataset_id in ir_datasets.registry._registered:
        return
    base_path = ir_datasets.util.home_path() / base_id
    dlc = DownloadConfig.context(base_id, base_path)
    dlc._contents[cache_path] = dlc_contents
    ir_dataset = ir_datasets.load(f"{base_id}/{split_id}")
    collection = ir_dataset.docs_handler()
    queries = ir_dataset.queries_handler()
    qrels = ir_dataset.qrels_handler()
    _dlc = dlc[cache_path]
    if extract:
        _dlc = GzipExtract(_dlc)
    constituent = ConstituentType(Cache(_dlc, base_path / split_id / file_name))
    dataset = Dataset(collection, queries, qrels, constituent)
    ir_datasets.registry.register(dataset_id, Dataset(dataset))


register_kd_docpairs()
register_colbert_docpairs()
register_rank_distillm()



from typing import Dict, List, Sequence, Tuple, Type

from transformers import BatchEncoding

from ..base import LightningIRTokenizer
from .config import CrossEncoderConfig


class CrossEncoderTokenizer(LightningIRTokenizer):

    config_class: Type[CrossEncoderConfig] = CrossEncoderConfig

    def __init__(self, *args, query_length: int = 32, doc_length: int = 512, **kwargs):
        super().__init__(*args, query_length=query_length, doc_length=doc_length, **kwargs)

    def truncate(self, text: Sequence[str], max_length: int) -> List[str]:
        return self.batch_decode(
            self(
                text,
                add_special_tokens=False,
                truncation=True,
                max_length=max_length,
                return_attention_mask=False,
                return_token_type_ids=False,
            ).input_ids
        )

    def expand_queries(self, queries: Sequence[str], num_docs: Sequence[int]) -> List[str]:
        return [query for query_idx, query in enumerate(queries) for _ in range(num_docs[query_idx])]

    def preprocess(
        self,
        queries: str | Sequence[str] | None,
        docs: str | Sequence[str] | None,
        num_docs: Sequence[int] | None,
    ) -> Tuple[str | Sequence[str], str | Sequence[str]]:
        if queries is None or docs is None:
            raise ValueError("Both queries and docs must be provided.")
        queries_is_string = isinstance(queries, str)
        docs_is_string = isinstance(docs, str)
        if queries_is_string != docs_is_string:
            raise ValueError("Queries and docs must be both lists or both strings.")
        if queries_is_string and docs_is_string:
            queries = [queries]
            docs = [docs]
        truncated_queries = self.truncate(queries, self.query_length)
        truncated_docs = self.truncate(docs, self.doc_length)
        if not queries_is_string:
            if num_docs is None:
                num_docs = [len(docs) // len(queries) for _ in range(len(queries))]
            expanded_queries = self.expand_queries(truncated_queries, num_docs)
            docs = truncated_docs
        else:
            expanded_queries = truncated_queries[0]
            docs = truncated_docs[0]
        return expanded_queries, docs

    def tokenize(
        self,
        queries: str | Sequence[str] | None = None,
        docs: str | Sequence[str] | None = None,
        num_docs: Sequence[int] | None = None,
        **kwargs,
    ) -> Dict[str, BatchEncoding]:
        expanded_queries, docs = self.preprocess(queries, docs, num_docs)
        return_tensors = kwargs.get("return_tensors", None)
        if return_tensors is not None:
            kwargs["pad_to_multiple_of"] = 8
        return {"encoding": self(expanded_queries, docs, **kwargs)}



import warnings
from itertools import islice
from pathlib import Path
from typing import Any, Dict, Iterator, Literal, Tuple

import ir_datasets
import numpy as np
import pandas as pd
import torch
from ir_datasets.formats import GenericDoc, GenericDocPair
from torch.distributed import get_rank, get_world_size
from torch.utils.data import Dataset, IterableDataset, get_worker_info

from .data import DocSample, QuerySample, RankSample
from .ir_datasets_utils import ScoredDocTuple

RUN_HEADER = ["query_id", "q0", "doc_id", "rank", "score", "system"]


class IRDataset:
    def __init__(self, dataset: str) -> None:
        super().__init__()
        if dataset in self.DASHED_DATASET_MAP:
            dataset = self.DASHED_DATASET_MAP[dataset]
        self.dataset = dataset
        try:
            self.ir_dataset = ir_datasets.load(dataset)
        except KeyError:
            self.ir_dataset = None
        self._queries = None
        self._docs = None
        self._qrels = None

    @property
    def DASHED_DATASET_MAP(self) -> Dict[str, str]:
        return {dataset.replace("/", "-"): dataset for dataset in ir_datasets.registry._registered}

    @property
    def queries(self) -> pd.Series:
        if self._queries is None:
            if self.ir_dataset is None:
                raise ValueError(f"Unable to find dataset {self.dataset} in ir-datasets")
            queries_iter = self.ir_dataset.queries_iter()
            self._queries = pd.Series(
                {query.query_id: query.default_text() for query in queries_iter},
                name="text",
            )
            self._queries.index.name = "query_id"
        return self._queries

    @property
    def docs(self) -> ir_datasets.indices.Docstore | Dict[str, GenericDoc]:
        if self._docs is None:
            if self.ir_dataset is None:
                raise ValueError(f"Unable to find dataset {self.dataset} in ir-datasets")
            self._docs = self.ir_dataset.docs_store()
        return self._docs

    @property
    def qrels(self) -> pd.DataFrame | None:
        if self._qrels is not None:
            return self._qrels
        if self.ir_dataset is None:
            return None
        qrels = pd.DataFrame(self.ir_dataset.qrels_iter()).rename({"subtopic_id": "iteration"}, axis=1)
        if "iteration" not in qrels.columns:
            qrels["iteration"] = 0
        qrels = qrels.drop_duplicates(["query_id", "doc_id", "iteration"])
        qrels = qrels.set_index(["query_id", "doc_id", "iteration"]).unstack(level=-1)
        qrels = qrels.droplevel(0, axis=1)
        self._qrels = qrels
        return self._qrels

    @property
    def dataset_id(self) -> str:
        if self.ir_dataset is None:
            return self.dataset
        return self.ir_dataset.dataset_id()

    @property
    def docs_dataset_id(self) -> str:
        return ir_datasets.docs_parent_id(self.dataset_id)


class DataParallelIterableDataset(IterableDataset):
    # https://github.com/Lightning-AI/pytorch-lightning/issues/15734
    def __init__(self) -> None:
        super().__init__()
        # TODO add support for multi-gpu and multi-worker inference; currently
        # doesn't work
        worker_info = get_worker_info()
        num_workers = worker_info.num_workers if worker_info is not None else 1
        worker_id = worker_info.id if worker_info is not None else 0

        try:
            world_size = get_world_size()
            process_rank = get_rank()
        except (RuntimeError, ValueError):
            world_size = 1
            process_rank = 0

        self.num_replicas = num_workers * world_size
        self.rank = process_rank * num_workers + worker_id


class QueryDataset(IRDataset, DataParallelIterableDataset):
    def __init__(self, query_dataset: str, num_queries: int | None = None) -> None:
        super().__init__(query_dataset)
        super(IRDataset, self).__init__()
        self.num_queries = num_queries

    def __len__(self) -> int:
        # TODO fix len for multi-gpu and multi-worker inference
        return self.num_queries or self.ir_dataset.queries_count()

    def __iter__(self) -> Iterator[QuerySample]:
        start = self.rank
        stop = self.num_queries
        step = self.num_replicas
        for sample in islice(self.ir_dataset.queries_iter(), start, stop, step):
            query_sample = QuerySample.from_ir_dataset_sample(sample)
            if self.qrels is not None:
                qrels = (
                    self.qrels.loc[[query_sample.query_id]]
                    .stack()
                    .rename("relevance")
                    .astype(int)
                    .reset_index()
                    .to_dict(orient="records")
                )
                query_sample.qrels = qrels
            yield query_sample


class DocDataset(IRDataset, DataParallelIterableDataset):
    def __init__(self, doc_dataset: str, num_docs: int | None = None) -> None:
        super().__init__(doc_dataset)
        super(IRDataset, self).__init__()
        self.num_docs = num_docs

    def __len__(self) -> int:
        # TODO fix len for multi-gpu and multi-worker inference
        return self.num_docs or self.ir_dataset.docs_count()

    def __iter__(self) -> Iterator[DocSample]:
        start = self.rank
        stop = self.num_docs
        step = self.num_replicas
        for sample in islice(self.ir_dataset.docs_iter(), start, stop, step):
            yield DocSample.from_ir_dataset_sample(sample)


class Sampler:

    @staticmethod
    def single_relevant(group: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        relevance = group.filter(like="relevance").max(axis=1).fillna(0)
        relevant = group.loc[relevance.gt(0)].sample(1)
        non_relevant_bool = relevance.eq(0) & ~group["rank"].isna()
        num_non_relevant = non_relevant_bool.sum()
        sample_non_relevant = min(sample_size - 1, num_non_relevant)
        non_relevant = group.loc[non_relevant_bool].sample(sample_non_relevant)
        return pd.concat([relevant, non_relevant])

    @staticmethod
    def top(group: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        return group.head(sample_size)

    @staticmethod
    def top_and_random(group: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        top_size = sample_size // 2
        random_size = sample_size - top_size
        top = group.head(top_size)
        random = group.iloc[top_size:].sample(random_size)
        return pd.concat([top, random])

    @staticmethod
    def random(group: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        return group.sample(sample_size)

    @staticmethod
    def log_random(group: pd.DataFrame, sample_size: int) -> pd.DataFrame:
        weights = 1 / np.log1p(group["rank"])
        weights[weights.isna()] = weights.min()
        return group.sample(sample_size, weights=weights)

    @staticmethod
    def sample(
        df: pd.DataFrame,
        sample_size: int,
        sampling_strategy: Literal["single_relevant", "top", "random", "log_random", "top_and_random"],
    ) -> pd.DataFrame:
        if sample_size == -1:
            return df
        if hasattr(Sampler, sampling_strategy):
            return getattr(Sampler, sampling_strategy)(df, sample_size)
        raise ValueError("Invalid sampling strategy.")


class RunDataset(IRDataset, Dataset):
    def __init__(
        self,
        run_path_or_id: Path | str,
        depth: int = -1,
        sample_size: int = -1,
        sampling_strategy: Literal["single_relevant", "top", "random", "log_random", "top_and_random"] = "top",
        targets: Literal["relevance", "subtopic_relevance", "rank", "score"] | None = None,
        normalize_targets: bool = False,
        add_non_retrieved_docs: bool = False,
    ) -> None:
        self.run_path = None
        if Path(run_path_or_id).is_file():
            self.run_path = Path(run_path_or_id)
            dataset = self.run_path.name.split(".")[0].split("__")[-1]
        else:
            dataset = str(run_path_or_id)
        super().__init__(dataset)
        self.depth = depth
        self.sample_size = sample_size
        self.sampling_strategy = sampling_strategy
        self.targets = targets
        self.normalize_targets = normalize_targets

        if self.sampling_strategy == "top" and self.sample_size > self.depth:
            warnings.warn(
                "Sample size is greater than depth and top sampling strategy is used. "
                "This can cause documents to be sampled that are not contained "
                "in the run file, but that are present in the qrels."
            )

        self.run = self.load_run()
        self.run = self.run.drop_duplicates(["query_id", "doc_id"])

        if self.qrels is not None:
            run_query_ids = pd.Index(self.run["query_id"].drop_duplicates())
            qrels_query_ids = self.qrels.index.get_level_values("query_id").unique()
            query_ids = run_query_ids.intersection(qrels_query_ids)
            if len(run_query_ids.difference(qrels_query_ids)):
                self.run = self.run[self.run["query_id"].isin(query_ids)]
            # outer join if docs are from ir_datasets else only keep docs in run
            how = "left"
            if self._docs is None and add_non_retrieved_docs:
                how = "outer"
            self.run = self.run.merge(
                self.qrels.loc[pd.IndexSlice[query_ids, :]].add_prefix("relevance_", axis=1),
                on=["query_id", "doc_id"],
                how=how,
            )

        if self.sample_size != -1:
            num_docs_per_query = self.run.groupby("query_id").transform("size")
            self.run = self.run[num_docs_per_query >= self.sample_size]

        self.run = self.run.sort_values(["query_id", "rank"])
        self.run_groups = self.run.groupby("query_id")
        self.query_ids = list(self.run_groups.groups.keys())

        if self.depth != -1 and self.run["rank"].max() < self.depth:
            warnings.warn("Depth is greater than the maximum rank in the run file.")

    @staticmethod
    def load_csv(path: Path) -> pd.DataFrame:
        return pd.read_csv(
            path,
            sep=r"\s+",
            header=None,
            names=RUN_HEADER,
            usecols=[0, 2, 3, 4],
            dtype={"query_id": str, "doc_id": str},
        )

    @staticmethod
    def load_parquet(path: Path) -> pd.DataFrame:
        return pd.read_parquet(path).rename(
            {
                "qid": "query_id",
                "docid": "doc_id",
                "docno": "doc_id",
            },
            axis=1,
        )

    @staticmethod
    def load_json(path: Path) -> pd.DataFrame:
        kwargs: Dict[str, Any] = {}
        if ".jsonl" in path.suffixes:
            kwargs["lines"] = True
            kwargs["orient"] = "records"
        run = pd.read_json(
            path,
            **kwargs,
            dtype={
                "query_id": str,
                "qid": str,
                "doc_id": str,
                "docid": str,
                "docno": str,
            },
        )
        return run

    def _get_run_path(self) -> Path | None:
        run_path = self.run_path
        if run_path is None:
            if self.ir_dataset is None or not self.ir_dataset.has_scoreddocs():
                raise ValueError("Run file or dataset with scoreddocs required.")
            try:
                run_path = self.ir_dataset.scoreddocs_handler().scoreddocs_path()
            except NotImplementedError:
                pass
        return run_path

    def _clean_run(self, run: pd.DataFrame) -> pd.DataFrame:
        run = run.rename(
            {"qid": "query_id", "docid": "doc_id", "docno": "doc_id"},
            axis=1,
        )
        if "query" in run.columns:
            self._queries = run.drop_duplicates("query_id").set_index("query_id")["query"].rename("text")
            run = run.drop("query", axis=1)
        if "text" in run.columns:
            self._docs = run.set_index("doc_id")["text"].map(lambda x: GenericDoc("", x)).to_dict()
            run = run.drop("text", axis=1)
        if self.depth != -1:
            run = run[run["rank"] <= self.depth]
        dtypes = {"rank": np.int32}
        if "score" in run.columns:
            dtypes["score"] = np.float32
        run = run.astype(dtypes)
        return run

    def load_run(self) -> pd.DataFrame:

        suffix_load_map = {
            ".tsv": self.load_csv,
            ".run": self.load_csv,
            ".csv": self.load_csv,
            ".parquet": self.load_parquet,
            ".json": self.load_json,
            ".jsonl": self.load_json,
        }
        run = None

        # try loading run from file
        run_path = self._get_run_path()
        if run_path is not None:
            load_func = suffix_load_map.get(run_path.suffixes[0], None)
            if load_func is not None:
                try:
                    run = load_func(run_path)
                except Exception:
                    pass

        # try loading run from ir_datasets
        if run is None and self.ir_dataset is not None and self.ir_dataset.has_scoreddocs():
            run = pd.DataFrame(self.ir_dataset.scoreddocs_iter())
            run["rank"] = run.groupby("query_id")["score"].rank("first", ascending=False)
            run = run.sort_values(["query_id", "rank"])

        if run is None:
            raise ValueError("Invalid run file format.")

        run = self._clean_run(run)
        return run

    @property
    def qrels(self) -> pd.DataFrame | None:
        if self._qrels is not None:
            return self._qrels
        if "relevance" in self.run:
            qrels = self.run[["query_id", "doc_id", "relevance"]].copy()
            if "iteration" in self.run:
                qrels["iteration"] = self.run["iteration"]
            else:
                qrels["iteration"] = "0"
            self.run = self.run.drop(["relevance", "iteration"], axis=1, errors="ignore")
            qrels = qrels.drop_duplicates(["query_id", "doc_id", "iteration"])
            qrels = qrels.set_index(["query_id", "doc_id", "iteration"]).unstack(level=-1)
            qrels = qrels.droplevel(0, axis=1)
            self._qrels = qrels
            return self._qrels
        return super().qrels

    def __len__(self) -> int:
        return len(self.query_ids)

    def __getitem__(self, idx: int) -> RankSample:
        query_id = str(self.query_ids[idx])
        group = self.run_groups.get_group(query_id).copy()
        query = self.queries[query_id]
        group = Sampler.sample(group, self.sample_size, self.sampling_strategy)

        doc_ids = tuple(group["doc_id"])
        docs = tuple(self.docs.get(doc_id).default_text() for doc_id in doc_ids)

        targets = None
        if self.targets is not None:
            filtered = group.set_index("doc_id").loc[list(doc_ids)].filter(like=self.targets).fillna(0)
            if filtered.empty:
                raise ValueError(f"targets `{self.targets}` not found in run file")
            targets = torch.from_numpy(filtered.values)
            if self.targets == "rank":
                # invert ranks to be higher is better (necessary for loss functions)
                targets = self.depth - targets + 1
            if self.normalize_targets:
                targets_min = targets.min()
                targets_max = targets.max()
                targets = (targets - targets_min) / (targets_max - targets_min)
        qrels = None
        if self.qrels is not None:
            qrels = (
                self.qrels.loc[[query_id]]
                .stack()
                .rename("relevance")
                .astype(int)
                .reset_index()
                .to_dict(orient="records")
            )
        return RankSample(query_id, query, doc_ids, docs, targets, qrels)


class TupleDataset(IRDataset, IterableDataset):
    def __init__(
        self,
        tuples_dataset: str,
        targets: Literal["order", "score"] = "order",
        num_docs: int | None = None,
    ) -> None:
        super().__init__(tuples_dataset)
        super(IRDataset, self).__init__()
        self.targets = targets
        self.num_docs = num_docs

    def parse_sample(
        self, sample: ScoredDocTuple | GenericDocPair
    ) -> Tuple[Tuple[str, ...], Tuple[str, ...], Tuple[float, ...] | None]:
        if isinstance(sample, GenericDocPair):
            if self.targets == "score":
                raise ValueError("ScoredDocTuple required for score targets.")
            targets = (1.0, 0.0)
            doc_ids = (sample.doc_id_a, sample.doc_id_b)
        elif isinstance(sample, ScoredDocTuple):
            doc_ids = sample.doc_ids[: self.num_docs]
            if self.targets == "score":
                if sample.scores is None:
                    raise ValueError("tuples dataset does not contain scores")
                targets = sample.scores
            elif self.targets == "order":
                targets = tuple([1.0] + [0.0] * (sample.num_docs - 1))
            else:
                raise ValueError(f"invalid value for targets, got {self.targets}, " "expected one of (order, score)")
            targets = targets[: self.num_docs]
        else:
            raise ValueError("Invalid sample type.")
        docs = tuple(self.docs.get(doc_id).default_text() for doc_id in doc_ids)
        return doc_ids, docs, targets

    def __iter__(self) -> Iterator[RankSample]:
        for sample in self.ir_dataset.docpairs_iter():
            query_id = sample.query_id
            query = self.queries.loc[query_id]
            doc_ids, docs, targets = self.parse_sample(sample)
            if targets is not None:
                targets = torch.tensor(targets)
            yield RankSample(query_id, query, doc_ids, docs, targets)



from dataclasses import dataclass
from typing import Type

import torch
from transformers import BatchEncoding

from ..base import LightningIRModel, LightningIROutput
from ..base.model import _batch_encoding
from . import CrossEncoderConfig


@dataclass
class CrossEncoderOutput(LightningIROutput):
    embeddings: torch.Tensor | None = None


class CrossEncoderModel(LightningIRModel):
    config_class: Type[CrossEncoderConfig] = CrossEncoderConfig

    def __init__(self, config: CrossEncoderConfig, *args, **kwargs):
        super().__init__(config, *args, **kwargs)
        self.config: CrossEncoderConfig
        self.linear = torch.nn.Linear(config.hidden_size, 1, bias=config.linear_bias)

    @_batch_encoding
    def forward(self, encoding: BatchEncoding) -> CrossEncoderOutput:
        embeddings = self._backbone_forward(**encoding).last_hidden_state
        embeddings = self._pooling(
            embeddings, encoding.get("attention_mask", None), pooling_strategy=self.config.pooling_strategy
        )
        scores = self.linear(embeddings).view(-1)
        return CrossEncoderOutput(scores=scores, embeddings=embeddings)



from pathlib import Path

import torch
from huggingface_hub import hf_hub_download
from transformers.modeling_utils import load_state_dict

from ...base import LightningIRModelClassFactory
from ...bi_encoder.model import BiEncoderEmbedding, ScoringFunction
from ..col import ColModel
from .config import XTRConfig


class XTRScoringFunction(ScoringFunction):
    def __init__(self, config: XTRConfig) -> None:
        super().__init__(config)
        self.config: XTRConfig

    def compute_similarity(
        self, query_embeddings: BiEncoderEmbedding, doc_embeddings: BiEncoderEmbedding
    ) -> torch.Tensor:
        similarity = super().compute_similarity(query_embeddings, doc_embeddings)

        if self.training and self.xtr_token_retrieval_k is not None:
            pass
            # TODO implement simulated token retrieval

            # if not torch.all(num_docs == num_docs[0]):
            #     raise ValueError("XTR token retrieval does not support variable number of documents.")
            # query_embeddings = query_embeddings[:: num_docs[0]]
            # doc_embeddings = doc_embeddings.view(1, 1, -1, doc_embeddings.shape[-1])
            # ib_similarity = super().compute_similarity(
            #     query_embeddings,
            #     doc_embeddings,
            #     query_scoring_mask[:: num_docs[0]],
            #     doc_scoring_mask.view(1, -1),
            #     num_docs,
            # )
            # top_k_similarity = ib_similarity.topk(self.xtr_token_retrieval_k, dim=-1)
            # cut_off_similarity = top_k_similarity.values[..., [-1]].repeat_interleave(num_docs, dim=0)
            # if self.fill_strategy == "min":
            #     fill = cut_off_similarity.expand_as(similarity)[similarity < cut_off_similarity]
            # elif self.fill_strategy == "zero":
            #     fill = 0
            # similarity[similarity < cut_off_similarity] = fill
        return similarity

    # def aggregate(
    #     self,
    #     scores: torch.Tensor,
    #     mask: torch.Tensor,
    #     query_aggregation_function: Literal["max", "sum", "mean", "harmonic_mean"],
    # ) -> torch.Tensor:
    #     if self.training and self.normalization == "Z":
    #         # Z-normalization
    #         mask = mask & (scores != 0)
    #     return super().aggregate(scores, mask, query_aggregation_function)


class XTRModel(ColModel):
    config_class = XTRConfig

    def __init__(self, config: XTRConfig, *args, **kwargs) -> None:
        super().__init__(config)
        self.scoring_function = XTRScoringFunction(config)
        self.config: XTRConfig

    @classmethod
    def from_pretrained(cls, model_name_or_path: str | Path, *args, **kwargs) -> "XTRModel":
        try:
            hf_hub_download(repo_id=str(model_name_or_path), filename="2_Dense/pytorch_model.bin")
        except Exception:
            return super().from_pretrained(model_name_or_path, *args, **kwargs)
        finally:
            return cls.from_xtr_checkpoint(model_name_or_path)

    @classmethod
    def from_xtr_checkpoint(cls, model_name_or_path: Path | str) -> "XTRModel":
        from transformers import T5EncoderModel

        cls = LightningIRModelClassFactory(XTRConfig).from_backbone_class(T5EncoderModel)
        config = cls.config_class.from_pretrained(model_name_or_path)
        config.update(
            {
                "name_or_path": str(model_name_or_path),
                "similarity_function": "dot",
                "query_aggregation_function": "sum",
                "query_expansion": False,
                "doc_expansion": False,
                "doc_pooling_strategy": None,
                "doc_mask_scoring_tokens": None,
                "normalize": True,
                "sparsification": None,
                "add_marker_tokens": False,
                "embedding_dim": 128,
                "projection": "linear_no_bias",
            }
        )
        state_dict_path = hf_hub_download(repo_id=str(model_name_or_path), filename="model.safetensors")
        state_dict = load_state_dict(state_dict_path)
        linear_state_dict_path = hf_hub_download(repo_id=str(model_name_or_path), filename="2_Dense/pytorch_model.bin")
        linear_state_dict = load_state_dict(linear_state_dict_path)
        linear_state_dict["projection.weight"] = linear_state_dict.pop("linear.weight")
        state_dict["encoder.embed_tokens.weight"] = state_dict["shared.weight"]
        state_dict.update(linear_state_dict)
        model = cls(config=config)
        model.load_state_dict(state_dict)
        return model



from pathlib import Path
from typing import Sequence

import pandas as pd
import pytest
from _pytest.fixtures import SubRequest

from lightning_ir import BiEncoderModule, LightningIRDataModule, LightningIRModule, LightningIRTrainer, RunDataset
from lightning_ir.lightning_utils.callbacks import IndexCallback, ReRankCallback, SearchCallback
from lightning_ir.retrieve import (
    FaissFlatIndexConfig,
    FaissSearchConfig,
    SearchConfig,
    SparseIndexConfig,
    SparseSearchConfig,
)
from lightning_ir.retrieve.indexer import IndexConfig

from .conftest import DATA_DIR


@pytest.fixture(
    params=[FaissFlatIndexConfig(), SparseIndexConfig()],
    ids=["Faiss", "Sparse"],
)
def index_config(request: SubRequest) -> IndexConfig:
    return request.param


def run_datamodule(module: LightningIRModule, inference_datasets: Sequence[RunDataset]) -> LightningIRDataModule:
    datamodule = LightningIRDataModule(
        module=module,
        num_workers=0,
        inference_batch_size=2,
        inference_datasets=inference_datasets,
    )
    datamodule.setup(stage="test")
    return datamodule


# @pytest.mark.parametrize("devices", (1, 2))
def test_index_callback(
    tmp_path: Path,
    bi_encoder_module: BiEncoderModule,
    doc_datamodule: LightningIRDataModule,
    index_config: IndexConfig,
    # devices: int,
):
    index_dir = tmp_path / "index"
    index_callback = IndexCallback(index_dir, index_config)

    trainer = LightningIRTrainer(
        # devices=devices,
        logger=False,
        enable_checkpointing=False,
        callbacks=[index_callback],
    )
    trainer.test(bi_encoder_module, datamodule=doc_datamodule)

    assert doc_datamodule.inference_datasets is not None
    assert index_callback.indexer.num_embeddings and index_callback.indexer.num_docs
    assert index_callback.indexer.num_embeddings >= index_callback.indexer.num_docs

    assert (index_dir / "index.faiss").exists() or (index_dir / "index.pt").exists()
    assert (index_dir / "doc_ids.txt").exists()
    doc_ids_path = index_dir / "doc_ids.txt"
    doc_ids = doc_ids_path.read_text().split()
    for idx, doc_id in enumerate(doc_ids):
        assert doc_id == f"doc_id_{idx+1}"
    assert (index_dir / "config.json").exists()


def get_index(
    bi_encoder_module: BiEncoderModule,
    doc_datamodule: LightningIRDataModule,
    search_config: SearchConfig,
) -> Path:
    if isinstance(search_config, FaissSearchConfig):
        index_type = "faiss"
        index_config = FaissFlatIndexConfig()
    elif isinstance(search_config, SparseSearchConfig):
        index_type = "sparse"
        index_config = SparseIndexConfig()
    else:
        raise ValueError("Unknown search_config type")
    index_dir = DATA_DIR / "indexes" / f"{index_type}-{bi_encoder_module.config.similarity_function}"
    if index_dir.exists():
        return index_dir / "lightning-ir"

    index_callback = IndexCallback(index_dir, index_config)

    trainer = LightningIRTrainer(
        logger=False,
        enable_checkpointing=False,
        callbacks=[index_callback],
    )
    trainer.test(bi_encoder_module, datamodule=doc_datamodule)
    return index_dir / "lightning-ir"


@pytest.mark.parametrize(
    "search_config",
    (
        FaissSearchConfig(k=5, imputation_strategy="min", candidate_k=10),
        FaissSearchConfig(k=5, imputation_strategy="gather", candidate_k=10),
        SparseSearchConfig(k=5),
    ),
    ids=["FaissMin", "FaissGather", "Sparse"],
)
def test_search_callback(
    tmp_path: Path,
    bi_encoder_module: BiEncoderModule,
    query_datamodule: LightningIRDataModule,
    doc_datamodule: LightningIRDataModule,
    search_config: SearchConfig,
):
    index_dir = get_index(bi_encoder_module, doc_datamodule, search_config)
    save_dir = tmp_path / "runs"
    search_callback = SearchCallback(index_dir, search_config, save_dir)

    trainer = LightningIRTrainer(
        logger=False,
        enable_checkpointing=False,
        callbacks=[search_callback],
        inference_mode=False,
    )
    trainer.test(bi_encoder_module, datamodule=query_datamodule)

    for dataloader in trainer.test_dataloaders:
        dataset = dataloader.dataset
        dataset_id = dataset.dataset_id.replace("/", "-")
        assert (save_dir / f"{dataset_id}.run").exists()
        run_df = pd.read_csv(
            save_dir / f"{dataset_id}.run",
            sep="\t",
            header=None,
            names=["query_id", "Q0", "doc_id", "rank", "score", "system"],
        )
        assert run_df["query_id"].nunique() == len(dataset)


def test_rerank_callback(tmp_path: Path, module: LightningIRModule, inference_datasets: Sequence[RunDataset]):
    datamodule = run_datamodule(module, inference_datasets)
    save_dir = tmp_path / "runs"
    rerank_callback = ReRankCallback(save_dir)
    trainer = LightningIRTrainer(logger=False, enable_checkpointing=False, callbacks=[rerank_callback])
    trainer.re_rank(module, datamodule)

    for dataloader in trainer.test_dataloaders:
        dataset = dataloader.dataset
        dataset_id = dataset.dataset_id.replace("/", "-")
        assert (save_dir / f"{dataset_id}.run").exists()
        run_df = pd.read_csv(
            save_dir / f"{dataset_id}.run",
            sep="\t",
            header=None,
            names=["query_id", "Q0", "doc_id", "rank", "score", "system"],
        )
        assert run_df["query_id"].nunique() == len(dataset)



from transformers import AutoConfig, AutoModel, AutoTokenizer

from ._register_external_models import _register_external_models
from .base import (
    LightningIRClassFactory,
    LightningIRConfig,
    LightningIRConfigClassFactory,
    LightningIRModel,
    LightningIRModelClassFactory,
    LightningIRModule,
    LightningIROutput,
    LightningIRTokenizer,
    LightningIRTokenizerClassFactory,
)
from .bi_encoder import (
    BiEncoderConfig,
    BiEncoderModel,
    BiEncoderModule,
    BiEncoderOutput,
    BiEncoderTokenizer,
    ScoringFunction,
)
from .cross_encoder import (
    CrossEncoderConfig,
    CrossEncoderModel,
    CrossEncoderModule,
    CrossEncoderOutput,
    CrossEncoderTokenizer,
)
from .data import (
    DocDataset,
    DocSample,
    IndexBatch,
    LightningIRDataModule,
    QueryDataset,
    QuerySample,
    RankSample,
    RunDataset,
    SearchBatch,
    TrainBatch,
    TupleDataset,
)
from .lightning_utils import (
    LR_SCHEDULERS,
    ConstantLRSchedulerWithLinearWarmup,
    GenericConstantSchedulerWithLinearWarmup,
    GenericConstantSchedulerWithQuadraticWarmup,
    GenericLinearSchedulerWithLinearWarmup,
    IndexCallback,
    LinearLRSchedulerWithLinearWarmup,
    RankCallback,
    ReRankCallback,
    SearchCallback,
    WarmupLRScheduler,
)
from .loss import (
    ApproxMRR,
    ApproxNDCG,
    ApproxRankMSE,
    ConstantMarginMSE,
    FLOPSRegularization,
    InBatchCrossEntropy,
    KLDivergence,
    L1Regularization,
    L2Regularization,
    LocalizedContrastiveEstimation,
    RankNet,
    SupervisedMarginMSE,
)
from .main import LightningIRTrainer
from .models import (
    ColConfig,
    ColModel,
    SpladeConfig,
    SpladeModel,
    T5CrossEncoderConfig,
    T5CrossEncoderModel,
    T5CrossEncoderTokenizer,
    XTRConfig,
    XTRModel,
)
from .retrieve import (
    FaissFlatIndexConfig,
    FaissFlatIndexer,
    FaissIVFIndexConfig,
    FaissIVFIndexer,
    FaissIVFPQIndexConfig,
    FaissIVFPQIndexer,
    FaissSearchConfig,
    FaissSearcher,
    IndexConfig,
    Indexer,
    SearchConfig,
    Searcher,
    SparseIndexConfig,
    SparseIndexer,
    SparseSearchConfig,
    SparseSearcher,
)

AutoConfig.register(BiEncoderConfig.model_type, BiEncoderConfig)
AutoModel.register(BiEncoderConfig, BiEncoderModel)
AutoTokenizer.register(BiEncoderConfig, BiEncoderTokenizer)
AutoConfig.register(CrossEncoderConfig.model_type, CrossEncoderConfig)
AutoModel.register(CrossEncoderConfig, CrossEncoderModel)
AutoTokenizer.register(CrossEncoderConfig, CrossEncoderTokenizer)
AutoConfig.register(ColConfig.model_type, ColConfig)
AutoModel.register(ColConfig, ColModel)
AutoTokenizer.register(ColConfig, BiEncoderTokenizer)
AutoConfig.register(SpladeConfig.model_type, SpladeConfig)
AutoModel.register(SpladeConfig, SpladeModel)
AutoTokenizer.register(SpladeConfig, BiEncoderTokenizer)
AutoConfig.register(T5CrossEncoderConfig.model_type, T5CrossEncoderConfig)
AutoModel.register(T5CrossEncoderConfig, T5CrossEncoderModel)
AutoTokenizer.register(T5CrossEncoderConfig, T5CrossEncoderTokenizer)
AutoConfig.register(XTRConfig.model_type, XTRConfig)
AutoModel.register(XTRConfig, XTRModel)
AutoTokenizer.register(XTRConfig, BiEncoderTokenizer)

_register_external_models()

__version__ = "0.0.1"

__all__ = [
    "ApproxMRR",
    "ApproxNDCG",
    "ApproxRankMSE",
    "BiEncoderConfig",
    "BiEncoderModel",
    "BiEncoderModule",
    "BiEncoderOutput",
    "BiEncoderTokenizer",
    "ColConfig",
    "ConstantLRSchedulerWithLinearWarmup",
    "ConstantMarginMSE",
    "CrossEncoderConfig",
    "CrossEncoderModel",
    "CrossEncoderModule",
    "CrossEncoderOutput",
    "CrossEncoderTokenizer",
    "DocDataset",
    "DocSample",
    "FaissFlatIndexConfig",
    "FaissFlatIndexer",
    "FaissIVFIndexConfig",
    "FaissIVFIndexer",
    "FaissIVFPQIndexConfig",
    "FaissIVFPQIndexer",
    "FaissSearchConfig",
    "FaissSearcher",
    "FLOPSRegularization",
    "GenericConstantSchedulerWithLinearWarmup",
    "GenericConstantSchedulerWithQuadraticWarmup",
    "GenericLinearSchedulerWithLinearWarmup",
    "InBatchCrossEntropy",
    "IndexBatch",
    "IndexCallback",
    "IndexConfig",
    "Indexer",
    "KLDivergence",
    "L1Regularization",
    "L2Regularization",
    "LightningIRClassFactory",
    "LightningIRConfig",
    "LightningIRConfigClassFactory",
    "LightningIRDataModule",
    "LightningIRModel",
    "LightningIRModelClassFactory",
    "LightningIRModule",
    "LightningIROutput",
    "LightningIRTokenizer",
    "LightningIRTokenizerClassFactory",
    "LightningIRTrainer",
    "LinearLRSchedulerWithLinearWarmup",
    "LocalizedContrastiveEstimation",
    "LR_SCHEDULERS",
    "QueryDataset",
    "QuerySample",
    "RankCallback",
    "RankNet",
    "RankSample",
    "ReRankCallback",
    "RunDataset",
    "ScoringFunction",
    "SearchBatch",
    "SearchCallback",
    "SearchConfig",
    "Searcher",
    "SparseIndexConfig",
    "SparseIndexer",
    "SparseSearchConfig",
    "SparseSearcher",
    "SupervisedMarginMSE",
    "T5CrossEncoderConfig",
    "T5CrossEncoderModel",
    "T5CrossEncoderTokenizer",
    "TrainBatch",
    "TupleDataset",
    "WarmupLRScheduler",
    "XTRConfig",
    "XTRModel",
]



from typing import Literal

from ...cross_encoder.config import CrossEncoderConfig


class T5CrossEncoderConfig(CrossEncoderConfig):

    model_type = "encoder-decoder-cross-encoder"

    TOKENIZER_ARGS = CrossEncoderConfig.TOKENIZER_ARGS.union({"decoder_strategy"})
    ADDED_ARGS = CrossEncoderConfig.ADDED_ARGS.union(TOKENIZER_ARGS)

    def __init__(
        self,
        query_length: int = 32,
        doc_length: int = 512,
        decoder_strategy: Literal["mono", "rank"] = "mono",
        **kwargs,
    ) -> None:
        kwargs["pooling_strategy"] = "first"
        super().__init__(query_length=query_length, doc_length=doc_length, **kwargs)
        self.decoder_strategy = decoder_strategy



from pathlib import Path
from typing import List, Sequence, Tuple

import torch

from ..base import LightningIRModule
from ..data import IndexBatch, RankBatch, SearchBatch, TrainBatch
from ..loss.loss import EmbeddingLossFunction, InBatchLossFunction, LossFunction, ScoringLossFunction
from ..retrieve import SearchConfig, Searcher
from .config import BiEncoderConfig
from .model import BiEncoderEmbedding, BiEncoderModel, BiEncoderOutput
from .tokenizer import BiEncoderTokenizer


class BiEncoderModule(LightningIRModule):
    def __init__(
        self,
        model_name_or_path: str | None = None,
        config: BiEncoderConfig | None = None,
        model: BiEncoderModel | None = None,
        loss_functions: Sequence[LossFunction | Tuple[LossFunction, float]] | None = None,
        evaluation_metrics: Sequence[str] | None = None,
        index_dir: Path | None = None,
        search_config: SearchConfig | None = None,
    ):
        super().__init__(model_name_or_path, config, model, loss_functions, evaluation_metrics)
        self.model: BiEncoderModel
        self.config: BiEncoderConfig
        self.tokenizer: BiEncoderTokenizer
        self.scoring_function = self.model.scoring_function
        if self.config.add_marker_tokens and len(self.tokenizer) > self.config.vocab_size:
            self.model.resize_token_embeddings(len(self.tokenizer), 8)
        self._searcher = None
        self.search_config = search_config
        self.index_dir = index_dir

    @property
    def searcher(self) -> Searcher | None:
        return self._searcher

    @searcher.setter
    def searcher(self, searcher: Searcher):
        self._searcher = searcher

    def on_test_start(self) -> None:
        if self.search_config is not None and self.index_dir is not None:
            self.searcher = self.search_config.search_class(self.index_dir, self.search_config, self)
        return super().on_test_start()

    def forward(self, batch: RankBatch | IndexBatch | SearchBatch) -> BiEncoderOutput:
        queries = getattr(batch, "queries", None)
        docs = getattr(batch, "docs", None)
        num_docs = None
        if isinstance(batch, RankBatch):
            num_docs = None if docs is None else [len(d) for d in docs]
            docs = [d for nested in docs for d in nested] if docs is not None else None
        encodings = self.prepare_input(queries, docs, num_docs)

        if not encodings:
            raise ValueError("No encodings were generated.")
        output = self.model.forward(
            encodings.get("query_encoding", None), encodings.get("doc_encoding", None), num_docs
        )
        if isinstance(batch, SearchBatch) and self.searcher is not None:
            scores, doc_ids, num_docs = self.searcher.search(output)
            output.scores = scores
            cum_num_docs = [0] + [sum(num_docs[: i + 1]) for i in range(len(num_docs))]
            doc_ids = tuple(tuple(doc_ids[cum_num_docs[i] : cum_num_docs[i + 1]]) for i in range(len(num_docs)))
            batch.doc_ids = doc_ids
        return output

    def score(self, queries: Sequence[str] | str, docs: Sequence[Sequence[str]] | Sequence[str]) -> BiEncoderOutput:
        return super().score(queries, docs)

    def compute_losses(self, batch: TrainBatch) -> List[torch.Tensor]:
        if self.loss_functions is None:
            raise ValueError("Loss function is not set")
        output = self.forward(batch)

        scores = output.scores
        query_embeddings = output.query_embeddings
        doc_embeddings = output.doc_embeddings
        if batch.targets is None or query_embeddings is None or doc_embeddings is None or scores is None:
            raise ValueError(
                "targets, scores, query_embeddings, and doc_embeddings must be set in " "the output and batch"
            )

        num_queries = len(batch.queries)
        scores = scores.view(num_queries, -1)
        targets = batch.targets.view(*scores.shape, -1)
        losses = []
        for loss_function, _ in self.loss_functions:
            if isinstance(loss_function, InBatchLossFunction):
                pos_idcs, neg_idcs = loss_function.get_ib_idcs(*scores.shape)
                ib_doc_embeddings = self.get_ib_doc_embeddings(doc_embeddings, pos_idcs, neg_idcs, num_queries)
                ib_scores = self.model.score(query_embeddings, ib_doc_embeddings)
                ib_scores = ib_scores.view(num_queries, -1)
                losses.append(loss_function.compute_loss(ib_scores))
            elif isinstance(loss_function, EmbeddingLossFunction):
                losses.append(loss_function.compute_loss(query_embeddings.embeddings, doc_embeddings.embeddings))
            elif isinstance(loss_function, ScoringLossFunction):
                losses.append(loss_function.compute_loss(scores, targets))
            else:
                raise ValueError(f"Unknown loss function type {loss_function.__class__.__name__}")
        if self.config.sparsification is not None:
            query_num_nonzero = (
                torch.nonzero(query_embeddings.embeddings).shape[0] / query_embeddings.embeddings.shape[0]
            )
            doc_num_nonzero = torch.nonzero(doc_embeddings.embeddings).shape[0] / doc_embeddings.embeddings.shape[0]
            self.log("query_num_nonzero", query_num_nonzero)
            self.log("doc_num_nonzero", doc_num_nonzero)
        return losses

    def get_ib_doc_embeddings(
        self,
        embeddings: BiEncoderEmbedding,
        pos_idcs: torch.Tensor,
        neg_idcs: torch.Tensor,
        num_queries: int,
    ) -> BiEncoderEmbedding:
        _, seq_len, emb_dim = embeddings.embeddings.shape
        ib_embeddings = torch.cat(
            [
                embeddings.embeddings[pos_idcs].view(num_queries, -1, seq_len, emb_dim),
                embeddings.embeddings[neg_idcs].view(num_queries, -1, seq_len, emb_dim),
            ],
            dim=1,
        ).view(-1, seq_len, emb_dim)
        ib_scoring_mask = torch.cat(
            [
                embeddings.scoring_mask[pos_idcs].view(num_queries, -1, seq_len),
                embeddings.scoring_mask[neg_idcs].view(num_queries, -1, seq_len),
            ],
            dim=1,
        ).view(-1, seq_len)
        return BiEncoderEmbedding(ib_embeddings, ib_scoring_mask)

    def validation_step(
        self,
        batch: TrainBatch | IndexBatch | SearchBatch | RankBatch,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> BiEncoderOutput:
        if isinstance(batch, IndexBatch):
            return self.forward(batch)
        if isinstance(batch, (RankBatch, TrainBatch, SearchBatch)):
            return super().validation_step(batch, batch_idx, dataloader_idx)
        raise ValueError(f"Unknown batch type {type(batch)}")



from .loss import (
    ApproxMRR,
    ApproxNDCG,
    ApproxRankMSE,
    ConstantMarginMSE,
    FLOPSRegularization,
    InBatchCrossEntropy,
    KLDivergence,
    L1Regularization,
    L2Regularization,
    LocalizedContrastiveEstimation,
    RankNet,
    SupervisedMarginMSE,
)

__all__ = [
    "ApproxMRR",
    "ApproxNDCG",
    "ApproxRankMSE",
    "ConstantMarginMSE",
    "FLOPSRegularization",
    "InBatchCrossEntropy",
    "KLDivergence",
    "L1Regularization",
    "L2Regularization",
    "LocalizedContrastiveEstimation",
    "RankNet",
    "SupervisedMarginMSE",
]



from .class_factory import (
    LightningIRClassFactory,
    LightningIRConfigClassFactory,
    LightningIRModelClassFactory,
    LightningIRTokenizerClassFactory,
)
from .config import LightningIRConfig
from .external_model_hub import CHECKPOINT_MAPPING, POST_LOAD_CALLBACKS, STATE_DICT_KEY_MAPPING
from .model import LightningIRModel, LightningIROutput
from .module import LightningIRModule
from .tokenizer import LightningIRTokenizer

__all__ = [
    "CHECKPOINT_MAPPING",
    "LightningIRClassFactory",
    "LightningIRConfigClassFactory",
    "LightningIRConfig",
    "LightningIRModel",
    "LightningIRModelClassFactory",
    "LightningIRModule",
    "LightningIROutput",
    "LightningIRTokenizer",
    "LightningIRTokenizerClassFactory",
    "POST_LOAD_CALLBACKS",
    "STATE_DICT_KEY_MAPPING",
]
